﻿using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace UserGuide
{
    /// <summary>
    /// Помощник создания и удаления диспетчера для работы WPF приложения.
    /// </summary>
    public static class DispatcherHelper
    {
        private static TaskCompletionSource dispatcherIsShutdown;
        private static volatile Dispatcher dispatcher;

        /// <summary>
        /// Диспетчер.
        /// Может быть <c>null</c>, если не запущен.
        /// </summary>
        public static Dispatcher Dispatcher => dispatcher;

        /// <summary>
        /// Запускает диспетчер.
        /// </summary>
        public static Task<Dispatcher> RunAsync()
        {
            // Диспетчер уже запущен.
            if (dispatcher is not null)
            {
                return Task.FromResult(dispatcher);
            }
            TaskCompletionSource<Dispatcher> dispatcherInitialized = new();
            dispatcherIsShutdown = new();
            var thread = new Thread(() =>
            {
                SynchronizationContext.SetSynchronizationContext(
                    new DispatcherSynchronizationContext(dispatcher = Dispatcher.CurrentDispatcher));
                dispatcherInitialized.SetResult(dispatcher);

                try
                {
                    Dispatcher.Run();
                }
                finally
                {
                    dispatcherIsShutdown.SetResult();
                }
            });

            thread.SetApartmentState(ApartmentState.STA);
            thread.Start();

            return dispatcherInitialized.Task;
        }

        /// <summary>
        /// Асинхронно завершает диспетчер.
        /// </summary>
        /// <returns>Задача для ожидания.</returns>
        public static async Task ShutdownAsync()
        {
            if (dispatcher is null)
            {
                return;
            }
            await dispatcher.InvokeAsync(dispatcher.InvokeShutdown);
            await dispatcherIsShutdown.Task;
            dispatcher = null;
        }
    }
}
