﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using UserGuide.Windows;
using OpenQA.Selenium;
using OpenQA.Selenium.Interactions;

namespace UserGuide.Extensions
{
    /// <summary>
    /// Класс расширений для <see cref="GuideProvider"/>.
    /// </summary>
    public static class GuideProviderExtensions
    {
        /// <summary>
        /// Получить построителя настроек окна подсказки.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <returns>Построитель настроек окна.</returns>
        public static TooltipOptionsBuilder GetTooltipOptionsBuilder(this GuideProvider provider) =>
            new TooltipOptionsBuilder(provider);

        #region Tooltip helpers

        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращённая задача ожидает, когда окно будет закрыто, и возвращает код закрытия окна.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="rect">Ограничивающий прямоугольник, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <param name="cancellationToken">Объект, посредством которого можно отменить асинхронную задачу.</param>
        /// <returns>Асинхронная задача с результатом закрытия окна.</returns>
        public static async Task<int> ShowTooltipAsync(
            this GuideProvider provider,
            string message,
            Rect rect,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator,
            CancellationToken cancellationToken = default)
        {
            return await 
                (await Tooltip.ShowAsync(
                    message,
                    rect,
                    configurator(provider.GetTooltipOptionsBuilder()).Build()))
                .WaitUntilWindowClosedAsync(cancellationToken);
        }
        
        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращённая задача ожидает, когда окно будет закрыто, и возвращает код закрытия окна.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="rect">Функция получения ограничивающего прямоугольника, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <param name="cancellationToken">Объект, посредством которого можно отменить асинхронную задачу.</param>
        /// <returns>Асинхронная задача с результатом закрытия окна.</returns>
        public static async Task<int> ShowTooltipAsync(
            this GuideProvider provider,
            string message,
            Func<Rect> rect,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator,
            CancellationToken cancellationToken = default)
        {
            return await
                (await Tooltip.ShowAsync(
                    message,
                    rect,
                    configurator(provider.GetTooltipOptionsBuilder()).Build()))
                .WaitUntilWindowClosedAsync(cancellationToken);
        }
        
        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращённая задача ожидает, когда окно будет закрыто, и возвращает код закрытия окна.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="element">Элемент, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <param name="cancellationToken">Объект, посредством которого можно отменить асинхронную задачу.</param>
        /// <returns>Асинхронная задача с результатом закрытия окна.</returns>
        public static async Task<int> ShowTooltipAsync(
            this GuideProvider provider,
            string message,
            IWebElement element,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator,
            CancellationToken cancellationToken = default)
        {
            return await 
                (await Tooltip.ShowAsync(
                    message,
                    element, 
                    configurator(provider.GetTooltipOptionsBuilder()).Build()))
                .WaitUntilWindowClosedAsync(cancellationToken);
        }

        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращает открытое окно. Для ожидания его закрытия заключите возвращённый задачей объект в блок <c>await using() { ... }</c>.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="rect">Ограничивающий прямоугольник, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <returns>Асинхронная задача с освобождаемым объектом, при освобождении которого будет закрыто окно.</returns>
        public static Task<Tooltip> ShowTooltipDisposableAsync(
            this GuideProvider provider,
            string message,
            Rect rect,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator) =>
            Tooltip.ShowAsync(
                message,
                rect,
                configurator(provider.GetTooltipOptionsBuilder()).Build());
        
        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращает открытое окно. Для ожидания его закрытия заключите возвращённый задачей объект в блок <c>await using() { ... }</c>.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="rect">Функция получения ограничивающего прямоугольника, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <returns>Асинхронная задача с освобождаемым объектом, при освобождении которого будет закрыто окно.</returns>
        public static Task<Tooltip> ShowTooltipDisposableAsync(
            this GuideProvider provider,
            string message,
            Func<Rect> rect,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator) =>
            Tooltip.ShowAsync(
                message,
                rect,
                configurator(provider.GetTooltipOptionsBuilder()).Build());
        
        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// Возвращает открытое окно. Для ожидания его закрытия заключите возвращённый задачей объект в блок <c>await using() { ... }</c>.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="message">Отображаемый текст.</param>
        /// <param name="element">Элемент, относительно которого нужно позиционировать окно.</param>
        /// <param name="configurator">Функция настроек окна.</param>
        /// <returns>Асинхронная задача с освобождаемым объектом, при освобождении которого будет закрыто окно.</returns>
        public static Task<Tooltip> ShowTooltipDisposableAsync(
            this GuideProvider provider,
            string message,
            IWebElement element,
            Func<TooltipOptionsBuilder, TooltipOptionsBuilder> configurator) =>
            Tooltip.ShowAsync(
                message,
                element,
                configurator(provider.GetTooltipOptionsBuilder()).Build());

        #endregion

        #region Yes No Dialogs

        /// <summary>
        /// Диалог в центре клиента со стандартными кнопками "Да", "Нет".
        /// Возвращённая задача дожидается закрытия окна и возвращает признак того, что окно было закрыто с результатом "Да".
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="question">Вопрос диалога.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом диалога.</returns>
        public static async Task<bool> YesNoDialogAsync(
            this GuideProvider provider,
            string question,
            CancellationToken cancellationToken = default)
        {
            return await provider.ShowTooltipAsync(question,
                provider.MainWindow,
                x => x
                    .UseInteractionColor()
                    .Place(Placement.Center)
                    .UseButtons(StandardButtons.Yes, StandardButtons.No),
                cancellationToken) == StandardButtons.YesCode;
        }

        /// <summary>
        /// Диалог в центре клиента со стандартными кнопками "Да", "Нет" и с текстом в формате Markdown.
        /// Возвращённая задача дожидается закрытия окна и возвращает признак того, что окно было закрыто с результатом "Да".
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="question">Вопрос диалога.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом диалога.</returns>
        public static async Task<bool> YesNoMarkdownDialogAsync(
            this GuideProvider provider,
            string question,
            CancellationToken cancellationToken = default)
        {
            return await provider.ShowTooltipAsync(question,
                provider.MainWindow,
                x => x
                    .UseInteractionColor()
                    .Place(Placement.Center)
                    .UseMarkdown()
                    .UseButtons(StandardButtons.Yes, StandardButtons.No),
                cancellationToken) == StandardButtons.YesCode;
        }

        #endregion

        #region Automation helpers
        
        /// <summary>
        /// Закрывает вкладку карточки.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="tab">Вкладка.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        public static async Task CloseTabAsync(
            this GuideProvider provider,
            IWebElement tab,
            CancellationToken cancellationToken = default)
        {
            await provider.HoverMouseAsync(tab, false, cancellationToken: cancellationToken);
            // Теперь находим кнопку, она должна появиться в дереве автоматизации.
            var tabCloseButton = tab.FindElement(By.Id("TabItem:CloseButton"));
            // И нажимаем на неё.
            await provider.ElementClickAsync(tabCloseButton, false, cancellationToken: cancellationToken);
        }

        /// <summary>
        /// Кликает по заданному элементу левой кнопкой мыши по алгоритму: наведение, задержка, клик, задержка.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="element">Элемент.</param>
        /// <param name="delay">Необходимо ли делать задержку после выполнения действия.</param>
        /// <param name="hover">Необходимо ли предварительно наводить курсор мыши на элемент.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        public static async Task ElementClickAsync(
            this GuideProvider provider,
            IWebElement element,
            bool delay = true,
            bool hover = true,
            CancellationToken cancellationToken = default)
        {
            if (hover)
            {
                await provider.HoverMouseAsync(element, delay, cancellationToken);
            }
            cancellationToken.ThrowIfCancellationRequested();
            element.Click();
            cancellationToken.ThrowIfCancellationRequested();
            if (delay)
            {
                await provider.DelayAsync(cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Кликает по заданному элементу правой кнопкой мыши по алгоритму: наведение, задержка, клик, задержка.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="element">Элемент.</param>
        /// <param name="delay">Необходимо ли делать задержку после выполнения действия.</param>
        /// <param name="hover">Необходимо ли предварительно наводить курсор мыши на элемент.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        public static async Task ElementRightClickAsync(
            this GuideProvider provider,
            IWebElement element,
            bool delay = true,
            bool hover = true,
            CancellationToken cancellationToken = default)
        {
            if (hover)
            {
                await provider.HoverMouseAsync(element, delay, cancellationToken);
            }
            cancellationToken.ThrowIfCancellationRequested();
            new Actions(provider.Driver)
                .ContextClick(element)
                .Build()
                .Perform();
            cancellationToken.ThrowIfCancellationRequested();
            if (delay)
            {
                await provider.DelayAsync(cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Дважды кликает по заданному элементу левой кнопкой мыши по алгоритму: наведение, задержка, двойной клик, задержка.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="element">Элемент.</param>
        /// <param name="delay">Необходимо ли делать задержку после выполнения действия.</param>
        /// <param name="hover">Необходимо ли предварительно наводить курсор мыши на элемент.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        public static async Task ElementDoubleClickAsync(
            this GuideProvider provider,
            IWebElement element,
            bool delay = true,
            bool hover = true,
            CancellationToken cancellationToken = default)
        {
            if (hover)
            {
                await provider.HoverMouseAsync(element, delay, cancellationToken);
            }
            cancellationToken.ThrowIfCancellationRequested();
            new Actions(provider.Driver)
                .DoubleClick(element)
                .Build()
                .Perform();
            cancellationToken.ThrowIfCancellationRequested();
            if (delay)
            {
                await provider.DelayAsync(cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Наводит мышь на указанный элемент с последующей задержкой.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="element">Элемент.</param>
        /// <param name="delay">Необходимо ли делать задержку после выполнения действия.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        public static async Task HoverMouseAsync(
            this GuideProvider provider,
            IWebElement element,
            bool delay = true,
            CancellationToken cancellationToken = default)
        {
            cancellationToken.ThrowIfCancellationRequested();
            new Actions(provider.Driver)
                .MoveToElement(element)
                .Build()
                .Perform();
            cancellationToken.ThrowIfCancellationRequested();
            if (delay)
            {
                await provider.DelayAsync(cancellationToken: cancellationToken);
            }
        }
        
        #endregion

        #region Delay helpers

        /// <summary>
        /// Метод выполнения ожидания в масштабе <paramref name="multiplier"/> стандартных интервалов задержки.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        /// <param name="multiplier">Задержка в долях от базовой задержки.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        /// <exception cref="ArgumentOutOfRangeException">Значение <c>multiplier</c> отрицательное.</exception>
        public static Task DelayAsync(
            this GuideProvider provider,
            double multiplier = 1,
            CancellationToken cancellationToken = default) =>
            multiplier < 0
            ? throw new ArgumentOutOfRangeException(nameof(multiplier), $"Передано недопустимое значение \"{multiplier}\". Значение должно быть больше или равно нулю.")
            : Task.Delay(provider.GeneralDelay * multiplier, cancellationToken);
        
        #endregion
    }
}
