﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using OpenQA.Selenium;

namespace UserGuide
{
    /// <summary>
    /// Помощники для обучения.
    /// </summary>
    public static class GuideHelper
    {
        #region Task execution helpers

        /// <summary>
        /// Выполнить действие с возможностью повторения.
        /// </summary>
        /// <param name="actionAsync">Основное действие.</param>
        /// <param name="replayPredicateAsync">Функция проверки необходимости повторения действия.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом выполнения исходной задачи.</returns>
        public static async Task<bool> RunActionWithReplayAsync(
            Func<CancellationToken, Task> actionAsync,
            Func<CancellationToken, Task<bool>> replayPredicateAsync,
            CancellationToken cancellationToken = default)
        {
            bool replay, isSuccessful;
            do
            {
                cancellationToken.ThrowIfCancellationRequested();

                try
                {
                    await actionAsync(cancellationToken);
                    isSuccessful = true;
                }
                catch (OperationCanceledException)
                {
                    throw;
                }
                catch (NotFoundException)
                {
                    isSuccessful = false;
                }
                catch (GuideDiscardException)
                {
                    isSuccessful = false;
                }

                cancellationToken.ThrowIfCancellationRequested();

                replay = !isSuccessful && await replayPredicateAsync(cancellationToken);
            }
            while (replay);
            return isSuccessful;
        }

        public static ValueTask ShowErrorAsync(Exception e, string message = null)
        {
            var dispatcher = DispatcherHelper.Dispatcher;
            if (dispatcher.CheckAccess())
            {
                MessageBox.Show(e.Message, message ?? "Зафиксирована ошибка сценария!", MessageBoxButton.OK, MessageBoxImage.Error);
                return ValueTask.CompletedTask;
            }

            return new(dispatcher.InvokeAsync(() =>
                MessageBox.Show(e.Message, message ?? "Зафиксирована ошибка сценария!", MessageBoxButton.OK, MessageBoxImage.Error)).Task);
        }

        #endregion
    }
}
