﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Controls;
using UserGuide.Extensions;
using UserGuide.Windows;
using OpenQA.Selenium;
using OpenQA.Selenium.Winium;

namespace UserGuide
{
    /// <summary>
    /// Класс с методами обучения пользователя.
    /// </summary>
    public class GuideProvider
    {
        #region Internals

        private record GuideDescriptor
        {
            public string Caption { get; init; }

            public Func<CancellationToken, Task> RunAsync { get; init; }
        }

        #endregion

        #region Private fields

        /// <summary>
        /// Настройки обучения.
        /// </summary>
        private readonly Options options;

        /// <summary>
        /// Драйвер Winium.
        /// </summary>
        private WiniumDriver driver;

        /// <summary>
        /// Главное окно TessaClient.
        /// </summary>
        private IWebElement mainWindow;

        /// <summary>
        /// Все известные и не отключенные сценарии.
        /// Индекс в массиве - числовое значение <see cref="GuideState"/>.
        /// </summary>
        private static readonly List<IGuide>[] KnownGuidesByState;

        /// <summary>
        /// Сценарии, подготовленные к запуску.
        /// Индекс в массиве - числовое значение <see cref="GuideState"/>.
        /// </summary>
        private readonly List<GuideDescriptor>[] guidesByState;

        /// <summary>
        /// Окно управления обучением (для пользователя).
        /// </summary>
        private TrainingControlWindow controlWindow;

        #endregion

        #region Ctor

        static GuideProvider()
        {
            KnownGuidesByState = new List<IGuide>[Enum.GetNames<GuideState>().Length];
            for (int i = 0; i < KnownGuidesByState.Length; i++)
            {
                KnownGuidesByState[i] = new();
            }

            DiscoverGuides();
        }

        /// <summary>
        /// Создаёт экземпляр класса для обучения пользователя по TessaClient.
        /// </summary>
        /// <param name="options">Настройки обучения.</param>
        public GuideProvider(Options options)
        {
            this.options = options;

            this.guidesByState = new List<GuideDescriptor>[KnownGuidesByState.Length];
            this.PrepareGuides();
        }

        #endregion

        #region Public Properties

        /// <summary>
        /// Интервал задержки.
        /// </summary>
        public TimeSpan GeneralDelay => this.options.GeneralDelay;

        /// <summary>
        /// Базовый интервал задержки для отображения диалога с подсказкой пользователю.
        /// Все диалоги должны использовать коэффициент масштабирования относительно этой величины.
        /// </summary>
        public TimeSpan DialogDelay => this.options.DialogDelay;

        /// <summary>
        /// Максимальное время ожидания выполнения действия пользователем.
        /// </summary>
        public TimeSpan MaxInteractionDelay => this.options.MaxInteractionDelay;

        /// <summary>
        /// Драйвер Winium.
        /// </summary>
        public WiniumDriver Driver => this.driver;

        /// <summary>
        /// Главное окно TessaClient.
        /// </summary>
        public IWebElement MainWindow => this.mainWindow;

        #endregion

        #region Public methods

        /// <summary>
        /// Начать обучение.
        /// </summary>
        /// <returns>Задача.</returns>
        public async Task RunAsync()
        {
            try
            {
                // Инициализируем драйвер и находим главное окно приложения.
                await this.InitializeAsync();

                try
                {
                    // Краткое обучение по подсказкам.
                    await RunTrainingChainAsync(this.GetGuideDescriptors(GuideState.Introduction), this.controlWindow.CancellationToken);
                }
                catch (OperationCanceledException)
                {
                    if (this.controlWindow.State == TrainingControlState.Exit)
                    {
                        // Нужно выйти, просто пробрасываем исключение дальше.
                        throw;
                    }
                    // Сбрасываем состояние
                    this.controlWindow.ResetState();
                }
                catch (Exception e)
                {
                    await GuideHelper.ShowErrorAsync(e);
                }

                // Главный цикл
                // Меню выбора разделов обучения.
                var (descriptors, actionsMap) = this.GetDescriptors();

                bool showMenu;
                do
                {
                    try
                    {
                        int choice = await this.ShowTooltipAsync("Выберите раздел обучения, который вы хотите пройти:",
                            this.mainWindow,
                            x => x
                                .UseInteractionColor()
                                .Place(Placement.Center)
                                .UseMarkdown()
                                .UseButtons(descriptors, Orientation.Vertical),
                            this.controlWindow.CancellationToken);

                        await GuideHelper.RunActionWithReplayAsync(actionsMap[choice], this.StandardPredicateAsync, this.controlWindow.CancellationToken)
                            .WaitAsync(this.controlWindow.CancellationToken);
                        showMenu = false;
                    }
                    catch (OperationCanceledException)
                    {
                        if (this.controlWindow.State == TrainingControlState.Exit)
                        {
                            // Нужно выйти, просто пробрасываем исключение дальше.
                            throw;
                        }
                        showMenu = true;
                        // Сбрасываем состояние.
                        this.controlWindow.ResetState();
                    }
                    catch (Exception e)
                    {
                        await GuideHelper.ShowErrorAsync(e);
                        showMenu = true;
                    }
                } while (showMenu);

                try
                {
                    this.controlWindow.HideMenuButton();
                    await RunTrainingChainAsync(this.GetGuideDescriptors(GuideState.Conclusion), this.controlWindow.CancellationToken);
                }
                catch (OperationCanceledException)
                {
                    if (this.controlWindow.State == TrainingControlState.Exit)
                    {
                        // Нужно выйти, просто пробрасываем исключение дальше.
                        throw;
                    }
                    // Сбрасываем состояние
                    this.controlWindow.ResetState();
                }
                catch (Exception e)
                {
                    await GuideHelper.ShowErrorAsync(e);
                }
            }
            catch (OperationCanceledException)
            {
                // ничего не делаем - выход по меню.
            }
            catch (Exception e)
            {
                await GuideHelper.ShowErrorAsync(e);
            }
            finally
            {
                if (this.controlWindow is not null)
                {
                    await this.controlWindow.CloseAsync();
                }
                if (this.driver is not null)
                {
                    this.driver.Quit();
                }
            }
        }

        /// <summary>
        /// Запустить указанный сценарий.
        /// </summary>
        /// <param name="guideType">Тип целевого сценария.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Задача.</returns>
        public Task RunGuideAsync(Type guideType, CancellationToken cancellationToken)
        {
            // Проверим, что этот тип соответствует типу сценария обучения.
            if (!typeof(IGuide).IsAssignableFrom(guideType))
            {
                return Task.CompletedTask;
            }
            // Теперь создадим его, инициализируем и запустим.
            var guide = (IGuide) Activator.CreateInstance(guideType);
            guide.Provider = this;
            return guide.StartAsync(cancellationToken);
        }

        /// <summary>
        /// Запустить указанный сценарий.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Задача.</returns>
        public Task RunGuideAsync<T>(CancellationToken cancellationToken)
            where T : IGuide =>
            this.RunGuideAsync(typeof(T), cancellationToken);

        /// <summary>
        /// Ожидает выполнения действия пользователем, в случае не выполнения действия генерирует исключение.
        /// </summary>
        /// <param name="predicate">Предикат, указывающий, что действие выполнено.</param>
        /// <param name="multiplier">Коэффициент ожидания от максимального ожидания действий пользователя.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом ожидания действия.</returns>
        public async Task WaitSuccessActionOrThrowAsync<T>(
            Func<Task<bool>> predicate,
            double multiplier = 1,
            CancellationToken cancellationToken = default) where T : Exception, new()
        {
            if (multiplier <= 0)
            {
                throw new ArgumentOutOfRangeException(nameof(multiplier), "Argument must be greater than 0.");
            }
            if (!await WaitActionAsync(predicate, this.MaxInteractionDelay.TotalMilliseconds * multiplier, cancellationToken))
            {
                throw new T();
            }
        }

        #endregion

        #region Private methods

        #region Initialization

        /// <summary>
        /// Инициализация драйвера подключения и поиск главного окна приложения. 
        /// </summary>
        private async Task InitializeAsync()
        {
            string driverPath = this.options.DriverPath?.Trim();
            string clientPath = this.options.ClientPath?.Trim();
            string userName = this.options.UserName?.Trim();
            string password = this.options.Password?.Trim();
            if (string.IsNullOrEmpty(password))
            {
                password = userName;
            }

            var options = new DesktopOptions()
            {
                ApplicationPath = clientPath,
                Arguments =
                    !string.IsNullOrEmpty(clientPath) && !string.IsNullOrEmpty(userName)
                        ? $"\"/u:{userName}\" \"/p:{password}\""
                        : string.Empty,
                DebugConnectToRunningApp = string.IsNullOrEmpty(clientPath)
            };

            try
            {
                if (!string.IsNullOrEmpty(driverPath))
                {
                    var service = WiniumDriverService.CreateDefaultService(
                        System.IO.Path.GetDirectoryName(driverPath),
                        System.IO.Path.GetFileName(driverPath));
                    service.EnableVerboseLogging = true;
                    service.ElementSearchTimeThreshold = TimeSpan.FromMilliseconds(100);

                    this.driver = new WiniumDriver(service, options);
                }
                else
                {
                    this.driver = new WiniumDriver(new Uri("http://localhost:9999"), options);
                }
            }
            catch (Exception e)
            {
                throw new Exception($"Не удалось подключится к Winium.\n---------------------------\n{e.Message}", e);
            }

            this.driver.Manage().Timeouts().ImplicitWait = TimeSpan.FromMilliseconds(100);

            if (!string.IsNullOrEmpty(clientPath))
            {
                // Если мы запускаем новый экземпляр приложения, нужно дождаться его запуска.
                if (!await WaitActionAsync(
                    async () => this.driver.FindElement(By.XPath("/*[@ClassName='Window' and starts-with(@Name, 'TessaClient')]")) is not null,
                    15_000, CancellationToken.None))
                {
                    throw new Exception("Не найдено окно TessaClient.");
                }
            }

            try
            {
                this.mainWindow = this.driver.FindElement(By.XPath("/*[@ClassName='Window' and starts-with(@Name, 'TessaClient')]"));
            }
            catch (Exception e)
            {
                // Так более понятно.
                throw new Exception("Не найдено окно TessaClient.", e);
            }

            var rect = this.MainWindow.GetRect();
            rect.Inflate(-10, -10);
            this.controlWindow = await TrainingControlWindow.ShowAsync(rect);
        }

        #endregion

        #region Guides helpers

        private static void DiscoverGuides()
        {
            var assembly = Assembly.GetExecutingAssembly();
            Type iGuideType = typeof(IGuide);
            foreach (var type in assembly.GetTypes())
            {
                // Проверяем, подходит ли этот тип нам или нет.
                if (!iGuideType.IsAssignableFrom(type) || type.IsInterface || type.IsAbstract || !type.IsPublic)
                {
                    continue;
                }
                // Создаём его экземпляр.
                var guide = (IGuide) Activator.CreateInstance(type);
                // Если он отключен, прекращаем рассмотрение.
                if (guide.State == GuideState.Disabled)
                {
                    continue;
                }
                // регистрируем в нужном списке.
                KnownGuidesByState[(int)guide.State].Add(guide);
            }
        }

        private void PrepareGuides()
        {
            for (int i = 0; i < KnownGuidesByState.Length; i++)
            {
                this.guidesByState[i] = KnownGuidesByState[i]
                    .OrderBy(x => x.Order)
                    .ThenBy(x => x.GetType().FullName)
                    .Select(x => new GuideDescriptor()
                    {
                        Caption = x.Caption,
                        RunAsync = async (ct) =>
                        {
                            x.Provider = this;
                            await x.StartAsync(ct);
                        }
                    })
                    .ToList();
            }

            // Для действий в Меню добавляем виртуальное действие: "Обучение целиком"
            var menuGuides = this.guidesByState[(int)GuideState.Menu];
            if (menuGuides.Count == 0)
            {
                throw new InvalidOperationException("Не задан ни один сценарий для обучения пользователя.");
            }
            //"Обучение целиком"
            menuGuides.Insert(0, new GuideDescriptor()
            {
                Caption = "Обучение целиком",
                RunAsync = this.RunWholeTrainingAsync,
            });
        }

        private (ButtonDescriptor[] descriptors, Func<CancellationToken, Task>[] actions)
            GetDescriptors()
        {
            var list = this.guidesByState[(int) GuideState.Menu];
            var descriptors = new ButtonDescriptor[list.Count];
            var actions = new Func<CancellationToken, Task>[list.Count];
            for (int i = 0; i < list.Count; ++i)
            {
                descriptors[i] = new ButtonDescriptor(list[i].Caption, i);
                actions[i] = list[i].RunAsync;
            }
            return (descriptors, actions);
        }

        #endregion

        #region Misc

        /// <summary>
        /// Всё обучение целиком.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Задача.</returns>
        private Task RunWholeTrainingAsync(CancellationToken cancellationToken) =>
            RunTrainingChainAsync(this.guidesByState[(int) GuideState.Menu].Skip(1), cancellationToken);

        /// <summary>
        /// Запустить на исполнение переданную цепочку действий.
        /// </summary>
        /// <param name="actions">Действия.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Задача.</returns>
        private static async Task RunTrainingChainAsync(
            IEnumerable<GuideDescriptor> actions,
            CancellationToken cancellationToken)
        {
            foreach (var action in actions)
            {
                await action.RunAsync(cancellationToken);
            }
        }

        /// <summary>
        /// Получить дескрипторы сценариев заданной группы.
        /// </summary>
        /// <param name="state">Группа сценариев.</param>
        /// <returns>Список (возможно пустой) сценариев.</returns>
        private IEnumerable<GuideDescriptor> GetGuideDescriptors(GuideState state) =>
            this.guidesByState[(int) state];

        /// <summary>
        /// Отображение стандартного вопроса о повторении обучения.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом диалога.</returns>
        private Task<bool> StandardPredicateAsync(
            CancellationToken cancellationToken) =>
            this.YesNoDialogAsync("При прохождении обучения произошла ошибка.\nХотите повторить?", cancellationToken);

        /// <summary>
        /// Ожидает выполнения действия пользователем.
        /// </summary>
        /// <param name="predicate">Предикат, указывающий, что действие выполнено.</param>
        /// <param name="waitThresholdMilliseconds">Максимальное время ожидания в миллисекундах.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача с результатом ожидания действия.</returns>
        private static async Task<bool> WaitActionAsync(
            Func<Task<bool>> predicate,
            double waitThresholdMilliseconds,
            CancellationToken cancellationToken)
        {
            bool isSuccessful = false;
            int delay = 1000;
            for (int i = 0; i < waitThresholdMilliseconds; i += delay)
            {
                try
                {
                    isSuccessful = await predicate();
                }
                catch (OperationCanceledException)
                {
                    throw;
                }
                catch (NotFoundException)
                {
                    isSuccessful = false;
                }
                catch (GuideDiscardException)
                {
                    isSuccessful = false;
                }

                if (isSuccessful)
                {
                    return true;
                }

                cancellationToken.ThrowIfCancellationRequested();

                await Task.Delay(delay, cancellationToken);
            }
            return isSuccessful;
        }

        #endregion

        #endregion
    }
}
