﻿using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using UserGuide.Extensions;
using UserGuide.Windows;
using OpenQA.Selenium;

namespace UserGuide.Guides
{
    /// <summary>
    /// Сценарий "Знакомство с интерфейсом".
    /// </summary>
    public sealed class ClientWindowGuide : Guide
    {
        /// <inheritdoc/>
        public override string Caption => "Информация о внешнем виде клиента";

        /// <inheritdoc/>
        public override int Order => 0;

        /// <inheritdoc/>
        public override async Task StartAsync(CancellationToken cancellationToken)
        {
            // Описание заголовка окна.
            await this.ExplainWindowHeaderAsync(cancellationToken);
            var tabControl = this.Provider.MainWindow.FindElement(By.Id("TabControl"));
            // Описание заголовка вкладок.
            await this.ExplainTabHeaderAsync(tabControl, cancellationToken);
            // Описание открытия и закрытия вкладки рабочего места.
            await this.ExplainOpenCloseWorkplaceAsync(tabControl, cancellationToken);
            // Гайд по рабочему месту.
            await this.ExplainWorkplaceElementsAsync(cancellationToken);
            // Описание панелей.
            await this.ExplainPanelsAsync(tabControl, cancellationToken);
        }

        #region Private methods

        /// <summary>
        /// Описание заголовка окна.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainWindowHeaderAsync(CancellationToken cancellationToken)
        {
            // Получим заголовок окна
            var header = this.Provider.MainWindow.FindElement(By.Id("DragWidget"));
            // Проверим, что операция не отменена
            cancellationToken.ThrowIfCancellationRequested();
            // Выделим заголовок
            await using (await Highlighter.ShowAsync(header))
            {
                // Отобразим подсказку в режиме автоматизации
                await this.Provider.ShowTooltipAsync("В верхнем правом углу приложения всегда отображается имя пользователя и логин, под которым выполнен вход в систему.",
                    () => // Передадим функцию вычисления ограничивающего прямоугольника
                    {
                        var rect = header.GetRect();
                        return new Rect(rect.Left + rect.Width / 2, rect.Top, rect.Width / 2, rect.Height);
                    },
                    x => x
                        .Place(Placement.Bottom, ArrowPlacement.Right) // Окно снизу элемента, стрелка выровнена по правому краю
                        .UseMarkdown() // Используется Markdown
                        .CloseAfter(6), // Закрыть окно через 6 секунд
                    cancellationToken);
                // Сделать дополнительную задержку перед следующим действием
                await this.Provider.DelayAsync(cancellationToken: cancellationToken);

                await this.Provider.ShowTooltipAsync("В верхнем левом углу приложения в скобках отображается псевдоним сервера, к которому сейчас подключено приложение. По данному тексту можно быстро определить, на тестовом или продуктивном сервере вы сейчас находитесь (например, `test`, `prod`). Это может быть полезно, если у вас есть доступ к нескольким серверам.",
                    () =>
                    {
                        var rect = header.GetRect();
                        return new Rect(rect.Left, rect.Top, rect.Width / 2, rect.Height);
                    },
                    x => x
                        .Place(Placement.Bottom, ArrowPlacement.Left)
                        .UseMarkdown()
                        .CloseAfter(12),
                    cancellationToken);
            }
        }

        /// <summary>
        /// Описание заголовка вкладок.
        /// </summary>
        /// <param name="tabControl">Объект с вкладками.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainTabHeaderAsync(IWebElement tabControl, CancellationToken cancellationToken)
        {
            // Получаем левую и правую кнопки открытия боковых панелей в заголовке вкладок
            var leftPanelButton = tabControl.FindElement(By.Id("TessaPanels:LeftPanelButton"));
            var rightPanelButton = tabControl.FindElement(By.Id("TessaPanels:RightPanelButton"));
            // Проверяем, что задача ещё не отменена
            cancellationToken.ThrowIfCancellationRequested();
            // Подсвечиваем вычисляемую область вкладок между левой и правой кнопками
            await using (await Highlighter.ShowAsync(() =>
            {
                Rect lpbRect = leftPanelButton.GetRect(), rpbRect = rightPanelButton.GetRect();
                return new Rect(lpbRect.Left, lpbRect.Top, rpbRect.Right - lpbRect.Left, rpbRect.Height);
            }))
            {
                // Отображаем окно с подсказкой
                await this.Provider.ShowTooltipAsync("Верхняя часть приложения - область вкладок. Тут отображаются открытые рабочие места и карточки.",
                    () => // Аналогично вычисляем область заголовка вкладок
                    {
                        Rect lpbRect = leftPanelButton.GetRect(), rpbRect = rightPanelButton.GetRect();
                        return new Rect(lpbRect.Left, lpbRect.Top, rpbRect.Right - lpbRect.Left, rpbRect.Height);
                    },
                    x => x.CloseAfter(6), // Закрываем через 6 секунд
                    cancellationToken);
            }
        }

        /// <summary>
        /// Описание открытия и закрытия рабочего места.
        /// </summary>
        /// <param name="tabControl">Объект с вкладками.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainOpenCloseWorkplaceAsync(IWebElement tabControl, CancellationToken cancellationToken)
        {
            var workplacesButton = tabControl.FindElement(By.Id("TessaWorkplaces:WorkplacesButton"));
            cancellationToken.ThrowIfCancellationRequested();

            await using (await Highlighter.ShowAsync(workplacesButton))
            {
                await this.Provider.ShowTooltipAsync("Посмотреть список доступных вам рабочих мест, а также открыть нужные можно через данную кнопку. Одно и то же рабочее место можно открыть несколько раз.",
                    workplacesButton,
                    x => x.CloseAfter(7),
                    cancellationToken);
            }

            // Все действия автоматизации, на которые может повлиять пользователь, вызываем в этой обёртке.
            // В случае ошибки, пользователя спросят, нужно ли продолжить.
            int userWorkplacesCount = this.Provider.MainWindow.FindElements(By.XPath("*[@AutomationId='TabControl']/*[@ClassName='TabItem' and @AutomationId='Workplace:c3d72683-f6c0-4766-a3d4-1fd9a7fe6827']")).Count;

            var workspaceOpened = await GuideHelper.RunActionWithReplayAsync(
                async ct =>
                {
                    // Открываем меню.
                    await this.Provider.ElementClickAsync(workplacesButton, cancellationToken: ct);

                    // В списке доступных рабочих мест находим пункт меню для рабочего места Пользователь.
                    var openUserWorkplaceMenuElement = this.Provider.MainWindow.FindElement(By.Name("$WorkplacesUser"));

                    // Открываем рабочее место.
                    await this.Provider.ElementClickAsync(openUserWorkplaceMenuElement, cancellationToken: ct);

                    if (this.Provider.MainWindow.FindElements(By.XPath("*[@AutomationId='TabControl']/*[@ClassName='TabItem' and @AutomationId='Workplace:c3d72683-f6c0-4766-a3d4-1fd9a7fe6827']")).Count <= userWorkplacesCount)
                    {
                        throw new GuideDiscardException("Не удалось открыть новое рабочее место пользователя.");
                    }
                },
                async (ct) =>
                    await this.Provider.YesNoDialogAsync("Не удалось открыть рабочее место \"Пользователь\".\nХотите повторить?", ct),
                cancellationToken);

            if (!workspaceOpened)
            {
                throw new GuideDiscardException("Не удалось открыть новое рабочее место пользователя.");
            }

            // Эта часть обучения запускается, только если выполнена предыдущая и пропускается в противном случае.
            var workspaceTab = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TabControl']/*[@ClassName='TabItem' and @AutomationId='Workplace:c3d72683-f6c0-4766-a3d4-1fd9a7fe6827'][last()]"));
            cancellationToken.ThrowIfCancellationRequested();

            await using (await Highlighter.ShowAsync(workspaceTab))
            {
                await this.Provider.ShowTooltipAsync("Закрыть вкладку (рабочее место или карточку) можно, подведя указатель мыши и нажав на колесо мыши, или нажатием на соответствующую иконку.",
                    workspaceTab,
                    x => x.CloseAfter(7),
                    cancellationToken);
            }

            // Здесь результат особо не важен. Можем просто пропустить.
            await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.Provider.CloseTabAsync(workspaceTab, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Не удалось закрыть рабочее место \"Пользователь\".\nХотите повторить?", ct),
                cancellationToken);
        }

        /// <summary>
        /// Описание элементов рабочего места.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainWorkplaceElementsAsync(CancellationToken cancellationToken)
        {
            var workspaceTab = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TabControl']/*[@ClassName='TabItem' and @AutomationId='Workplace:c3d72683-f6c0-4766-a3d4-1fd9a7fe6827'][last()]"));
            // Получаем содержимое рабочего места.
            var workplaceContent = workspaceTab.FindElement(By.Id("WorkplaceContent"));
            // Теперь получаем дерево рабочего места.
            var workplaceTree = workplaceContent.FindElement(By.Id("WorkplaceTree"));
            cancellationToken.ThrowIfCancellationRequested();

            await using (await Highlighter.ShowAsync(workplaceTree))
            {
                await this.Provider.ShowTooltipAsync("В левой части рабочего места располагается дерево доступных пользователю представлений. Существующие узлы дерева нельзя удалять или перемещать, но можно добавить новые - вывести узлы с преднастроенными фильтрами.",
                    workplaceTree,
                    x => x.Place(Placement.Right).CloseAfter(9),
                    cancellationToken);
            }

            // Получим первую таблицу данных первого открытого представления.
            var dataGrid = workplaceContent.FindElement(By.Id("DataGrid"));
            cancellationToken.ThrowIfCancellationRequested();

            await using (await Highlighter.ShowAsync(dataGrid))
            {
                await this.Provider.ShowTooltipAsync("В правой части рабочего места отображается содержимое выбранного в дереве представления. Представления - это не папки, в которые пользователь сам распределяет карточки, а некие наборы условий, по которым в том или ином представлении система отображает список строк (например, карточек или заданий).",
                    dataGrid,
                    x => x.CloseAfter(12),
                    cancellationToken);
            }
        }

        /// <summary>
        /// Описание панелей.
        /// </summary>
        /// <param name="tabControl">Объект с вкладками.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainPanelsAsync(IWebElement tabControl, CancellationToken cancellationToken)
        {
            // XPath использован для ускорения поиска элементов, т.к. известно в точности на каком уровне они находятся.
            var leftPanelButton = tabControl.FindElement(By.XPath("*[@AutomationId='TessaPanels:LeftPanelButton']"));
            var rightPanelButton = tabControl.FindElement(By.XPath("*[@AutomationId='TessaPanels:RightPanelButton']"));
            var leftSlider = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:LeftSlider']"));
            var rightSlider = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:RightSlider']"));
            var leftTopArea = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:LeftTopArea']"));
            var leftBottomArea = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:LeftBottomArea']"));
            var rightTopArea = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:RightTopArea']"));
            var rightBottomArea = this.Provider.MainWindow.FindElement(By.XPath("*[@AutomationId='TessaPanels:RightBottomArea']"));

            // Второй вариант поиска элементов закомментирован. Он найдёт все элементы, просто за более долгое время.
            //var leftPanelButton = tabControl.FindElement(By.Id("TessaPanels:LeftPanelButton"));
            //var rightPanelButton = tabControl.FindElement(By.Id("TessaPanels:RightPanelButton"));
            //var leftSlider = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:LeftSlider"));
            //var rightSlider = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:RightSlider"));
            //var leftTopArea = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:LeftTopArea"));
            //var leftBottomArea = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:LeftBottomArea"));
            //var rightTopArea = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:RightTopArea"));
            //var rightBottomArea = this.Provider.MainWindow.FindElement(By.Id("TessaPanels:RightBottomArea"));

            cancellationToken.ThrowIfCancellationRequested();

            await using (await Highlighter.ShowAsync(
                leftPanelButton, rightPanelButton, leftSlider, rightSlider,
                leftTopArea, leftBottomArea, rightTopArea, rightBottomArea))
            {
                await this.Provider.ShowTooltipAsync("Все кнопки для работы с системой скрыты в выпадающих панелях: правом и левом.  \nВызвать панель с кнопками можно, переместив указатель мыши в соответствующий край экрана, или нажав на кнопки `>` `<` в области вкладок.",
                    this.Provider.MainWindow,
                    x => x.Place(Placement.Center).CloseAfter(10),
                    cancellationToken);
            }

            // Описание левой панели.
            await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.ExplainLeftPanelAsync(leftSlider, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Не удалось открыть левую панель.\nХотите повторить?", ct),
                cancellationToken);

            // Описание правой панели.
            await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.ExplainRightPanelAsync(rightSlider, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Не удалось открыть правую панель.\nХотите повторить?", ct),
                cancellationToken);
        }

        /// <summary>
        /// Описание левой панели.
        /// </summary>
        /// <param name="sliderButton">Кнопка открытия панели.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainLeftPanelAsync(IWebElement sliderButton, CancellationToken cancellationToken)
        {
            // Откроем левую боковую панель.
            await this.Provider.ElementClickAsync(sliderButton, cancellationToken: cancellationToken);

            // Получим её содержимое.
            var commandPanel = this.Provider.MainWindow.FindElement(By.ClassName("Popup"));
            cancellationToken.ThrowIfCancellationRequested();

            await this.Provider.ShowTooltipAsync("Левая панель - контекстная. Состав кнопок тут всегда разный и зависит от того, где мы сейчас находимся. Это кнопки для работы с текущим объектом: представлением или карточкой.",
                commandPanel,
                x => x.Place(Placement.Right).CloseAfter(10),
                cancellationToken);

            // Найдём плитку "Выгрузить все данные".
            var tile = commandPanel.FindElement(By.Id("ViewAllDataExport"));
            await this.Provider.HoverMouseAsync(tile, delay: false, cancellationToken: cancellationToken);
            await this.Provider.ShowTooltipAsync("Например, кнопки для выгрузки информации из представления.",
                tile,
                x => x.Place(Placement.Right).CloseAfter(4),
                cancellationToken);

            await this.Provider.DelayAsync(cancellationToken: cancellationToken);
        }

        /// <summary>
        /// Описание правой панели.
        /// </summary>
        /// <param name="sliderButton">Кнопка открытия панели.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ExplainRightPanelAsync(IWebElement sliderButton, CancellationToken cancellationToken)
        {
            // Откроем левую боковую панель.
            await this.Provider.ElementClickAsync(sliderButton, cancellationToken: cancellationToken);

            // Получим её содержимое.
            var commandPanel = this.Provider.MainWindow.FindElement(By.ClassName("Popup"));
            cancellationToken.ThrowIfCancellationRequested();

            await this.Provider.ShowTooltipAsync("Правая панель - глобальная. Состав кнопок тут всегда одинаковый и не зависит от того, какое сейчас рабочее место или карточка открыты.",
                commandPanel,
                x => x.Place(Placement.Left).CloseAfter(6),
                cancellationToken);

            // Найдём плитку "О программе".
            var tile = commandPanel.FindElement(By.Id("About"));
            await this.Provider.HoverMouseAsync(tile, delay: false, cancellationToken: cancellationToken);
            await this.Provider.ShowTooltipAsync("На этой панели можно посмотреть информацию о приложении (версию и адрес сервера).",
                tile,
                x => x.Place(Placement.Left).CloseAfter(4),
                cancellationToken);

            // Найдём плитку "Создать карточку".
            tile = commandPanel.FindElement(By.Id("CreateCard"));
            await this.Provider.HoverMouseAsync(tile, delay: false, cancellationToken: cancellationToken);
            await this.Provider.ShowTooltipAsync("Создать новую карточку.",
               tile,
               x => x.Place(Placement.Left).CloseAfter(4),
               cancellationToken);

            // Найдём плитку "Мои настройки".
            tile = commandPanel.FindElement(By.Id("MySettings"));
            await this.Provider.HoverMouseAsync(tile, delay: false, cancellationToken: cancellationToken);
            await this.Provider.ShowTooltipAsync("Выполнить пользовательские настройки в системе.",
                tile,
                x => x.Place(Placement.Left).CloseAfter(4),
                cancellationToken);

            // Найдём плитку "Фон".
            tile = commandPanel.FindElement(By.Id("SelectWallpaper"));
            await this.Provider.HoverMouseAsync(tile, delay: false, cancellationToken: cancellationToken);
            await this.Provider.ShowTooltipAsync("Или сменить фон и тему приложения.",
               tile,
               x => x.Place(Placement.Left).CloseAfter(4),
               cancellationToken);

            // Отведём указатель от панели.
            await this.Provider.HoverMouseAsync(this.Provider.MainWindow, cancellationToken: cancellationToken);
        }

        #endregion
    }
}
