﻿using System.Threading;
using System.Threading.Tasks;
using UserGuide.Extensions;
using UserGuide.Windows;
using OpenQA.Selenium;

namespace UserGuide.Guides
{
    /// <summary>
    /// Сценарий работы с "Моими настройками".
    /// </summary>
    public sealed class MyPreferencesGuide : Guide
    {
        /// <inheritdoc/>
        public override string Caption => "\"Мои настройки\"";

        /// <inheritdoc/>
        public override int Order => 2;

        /// <inheritdoc/>
        public override async Task StartAsync(CancellationToken cancellationToken)
        {
            if (!await GuideHelper.RunActionWithReplayAsync(
                this.OpenMyPreferencesAsync,
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не открыли окно \"Мои настройки\".\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }

            // Найдём открытое окно "Мои настройки" и элемент с вкладками на форме карточки
            var myPreferences = this.Provider.MainWindow.FindElement(By.XPath("*[@ClassName='Window' and @Name='Мои настройки']"));
            var card = myPreferences.FindElement(By.Id("Card"));
            var tabControl = card.FindElement(By.Id("CardTabs"));
            // Проверим, что действие ещё не отменено
            cancellationToken.ThrowIfCancellationRequested();

            // Предлагаем пользователю установить флаг "Отключить затемнение для неактивного окна" с повтором выполнения действия.
            if (!await GuideHelper.RunActionWithReplayAsync(
                    async (ct) => await this.ChangeDisableWindowFadingAsync(tabControl, ct), // Действие проверки установки флага
                    async (ct) => await this.Provider.YesNoDialogAsync("Вы не установили флаг \"Отключить затемнение для неактивного окна\" на вкладке \"Мои настройки\".\nХотите повторить?", ct), // Предикат для повторного выполнения действия в случае неудачи
                    cancellationToken))
            {
                return;
            }

            // Далее идём к уведомлениям.
            if (!await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.OpenNotificationsTabAsync(tabControl, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не перешли на вкладку \"Уведомления\".\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }

            var tab = tabControl.FindElement(By.Name("$CardTypesTabs_NotificationSettings"));
            cancellationToken.ThrowIfCancellationRequested();

            await this.ShowNotificationHelpInfoAsync(tab, cancellationToken);

            if (!await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.AddNotificationRuleAsync(myPreferences, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не добавили требуемое правило уведомлений.\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }

            await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.ClosePreferencesAsync(myPreferences, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не закрыли окно \"Мои настройки\".\nХотите повторить?", ct),
                cancellationToken);
        }

        #region Private methods

        /// <summary>
        /// Ожидаем от пользователя открытия диалога "Мои настройки".
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task OpenMyPreferencesAsync(CancellationToken cancellationToken)
        {
            // Подсказка, что ожидается.
            await using (await this.Provider.ShowTooltipDisposableAsync(
                             "В правой панели системы вызовите окно настроек пользователя, нажав на кнопку \"Мои настройки\". Обратите внимание, что если около пункта выпадающего меню есть тонкая вертикальная полоска, то данный пункт меню также является кнопкой, на которую можно нажать.",
                             this.Provider.MainWindow,
                             x => x
                                 .UseInteractionColor()
                                 .Place(Placement.Center)))
            {
                // Проверяем выполнение с ожиданием.
                await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(async () =>
                        this.Provider.MainWindow.FindElement(By.XPath("*[@ClassName='Window' and @Name='Мои настройки']")) is not null,
                    cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Ожидаем от пользователя установку флага "Отключить затемнение для неактивного окна" на вкладке "Мои настройки".
        /// </summary>
        /// <param name="tabControl">Вкладки карточки.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ChangeDisableWindowFadingAsync(IWebElement tabControl, CancellationToken cancellationToken)
        {
            // Здесь прежде чем предлагать пользователю выполнить это действие, нужно проверить, что флаг действительно ещё не взведён.
            var tab = tabControl.FindElement(By.Id("PersonalRole"));
            cancellationToken.ThrowIfCancellationRequested();

            // Если вкладка выбрана, проверяем наличие флага.
            if (tab.Selected)
            {
                // Чтобы не показывать пользователю ненужные сообщения, проверяем флаг.
                var generalSettingsBlock = tab.FindElement(By.Id("GeneralSettings"));
                var disableWindowFadingControl = generalSettingsBlock.FindElement(By.Name("$CardTypes_Controls_DisableWindowFading"));
                var checkBox = disableWindowFadingControl.FindElement(By.Id("CheckBox"));
                // Требуемый флаг уже взведён, ничего делать не нужно.
                if (checkBox.Selected)
                {
                    return;
                }
                // Флаг не взведён, просим пользователя взвести его.
                await using (await this.Provider.ShowTooltipDisposableAsync(
                    "Для комфортного прохождения обучения рекомендуем выставить флаг \"Отключить затемнение для неактивного окна\".",
                    checkBox,
                    x => x.UseInteractionColor()))
                {
                    // Проверяем выполнение с ожиданием.
                    await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(
                        async () => checkBox.Selected,
                        cancellationToken: cancellationToken);
                }
                return;
            }
            // Эта часть выполнится, если мы на какой-то другой вкладке.
            // Выводим сообщение, что ожидается от пользователя
            await using (await this.Provider.ShowTooltipDisposableAsync(
                             "Для комфортного прохождения обучения рекомендуем выставить флаг \"Отключить затемнение для неактивного окна\".",
                             tabControl,
                             x => x.UseInteractionColor()))
            {
                // Проверяем выполнение с ожиданием.
                await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(async () =>
                {
                    // Находим флаг на вкладке
                    var generalSettingsBlock = tab.FindElement(By.Id("GeneralSettings"));
                    var disableWindowFadingControl = generalSettingsBlock.FindElement(By.Name("$CardTypes_Controls_DisableWindowFading"));
                    var checkBox = disableWindowFadingControl.FindElement(By.Id("CheckBox"));
                    cancellationToken.ThrowIfCancellationRequested();
                    // Проверяем, что вкладка выбрана и флаг установлен
                    return tab.Selected && checkBox.Selected;
                }, cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Ожидаем от пользователя переход на вкладку "Уведомления".
        /// </summary>
        /// <param name="tabControl">Вкладки карточки.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task OpenNotificationsTabAsync(IWebElement tabControl, CancellationToken cancellationToken)
        {
            var tab = tabControl.FindElement(By.Name("$CardTypesTabs_NotificationSettings"));
            cancellationToken.ThrowIfCancellationRequested();
            // Подсказка, что ожидается.
            await using (await this.Provider.ShowTooltipDisposableAsync(
                "На вкладке \"Уведомления\" можно настроить подписку на уведомления, получаемые на электронную почту, или отписаться от ненужных. Перейдите на данную вкладку.",
                tab,
                x => x.UseInteractionColor()))
            {
                // Проверяем выполнение с ожиданием.
                await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(
                    async () => tab.Selected,
                    cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Справка по вкладке "Уведомления".
        /// </summary>
        /// <param name="tab">Вкладка "Уведомления".</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ShowNotificationHelpInfoAsync(IWebElement tab, CancellationToken cancellationToken)
        {
            var block1 = tab.FindElement(By.Id("Block1"));
            var unsubscribeControl = block1.FindElement(By.Name("$CardTypes_Controls_UnsubscribedNotificationTypes"));
            var subscribeControl = block1.FindElement(By.Name("$CardTypes_Controls_SubscribedNotificationTypes"));

            await using (await Highlighter.ShowAsync(unsubscribeControl, subscribeControl))
            {
                await this.Provider.ShowTooltipAsync("В первых двух полях настраивается подписка/отписка по глобальным уведомлениям (т.е. уведомлениям, не по конкретной карточке/процессу, а системе в целом, например, еженочное уведомление со списком заданий).",
                    subscribeControl,
                    x => x.CloseAfter(6),
                    cancellationToken);
            }
        }

        /// <summary>
        /// Создание и настройка правила уведомления.
        /// </summary>
        /// <param name="window">Окно "Мои настройки".</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task AddNotificationRuleAsync(IWebElement window, CancellationToken cancellationToken)
        {
            if (!await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.OpenAddNotificationRuleDialogAsync(window, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не открыли окно \"Правило уведомления\".\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }

            if (!await GuideHelper.RunActionWithReplayAsync(
                this.ConfigureNotificationRuleAsync,
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не настроили правило.\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }

            if (!await GuideHelper.RunActionWithReplayAsync(
                async (ct) => await this.SaveNotificationRuleAsync(window, ct),
                async (ct) => await this.Provider.YesNoDialogAsync("Вы не сохранили правило.\nХотите повторить?", ct),
                cancellationToken))
            {
                return;
            }
        }

        /// <summary>
        /// Ожидаем от пользователя открытия окна "Правило уведомления".
        /// </summary>
        /// <param name="window">Окно "Мои настройки".</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task OpenAddNotificationRuleDialogAsync(IWebElement window, CancellationToken cancellationToken)
        {
            // Подсказка, что ожидается.
            await using (await this.Provider.ShowTooltipDisposableAsync(
                "Добавьте в таблицу \"Правила уведомлений\" новое правило.\nБудет открыт диалог \"Правило уведомления\".\nСамо правило мы настроим далее.",
                window,
                x => x.UseInteractionColor().Place(Placement.Center)))
            {
                // Проверяем выполнение с ожиданием.
                await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(async () =>
                    this.Provider.MainWindow.FindElement(By.XPath("*[@ClassName='Window' and @AutomationId='CardTableRowEditorDialog']")) is not null,
                    cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Ожидаем от пользователя настройки правила.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ConfigureNotificationRuleAsync(CancellationToken cancellationToken)
        {
            var window = this.Provider.MainWindow.FindElement(By.XPath("*[@ClassName='Window' and @AutomationId='CardTableRowEditorDialog']"));
            cancellationToken.ThrowIfCancellationRequested();
            // Подсказка, что ожидается.
            await this.Provider.ShowTooltipAsync(
                "Настройте правило, запрещающее получение уведомления (любого, какое выберите, или нескольких) по карточкам, где \"Тип документа\" = \"Входящий\" (в список условий добавьте тип условия по \"По типу документа\") и после, не закрывая окна \"Правило уведомления\", подтвердите выполнение настроек, нажав кнопку \"Готово\".",
                window,
                x => x
                    .UseInteractionColor()
                    .UseButtons(StandardButtons.Done)
                    .CloseAfterMaxInteractionDelay(1),
                cancellationToken);

            // Второй вариант, показывать окно пока не выполнено условие.
            //await using var _ = await this.Provider.ShowTooltipDisposableAsync(
            //    "Настройте правило, запрещающее получение уведомления (любого, какое выберите, или нескольких) по карточкам, где \"Тип документа\" = \"Входящий\" (в список условий добавьте тип условия по \"По типу документа\") и после, не закрывая окна \"Правило уведомления\", подтвердите выполнение настроек, нажав кнопку \"Готово\".",
            //    window,
            //    x => x.UseInteractionColor());

            var cardPanel = window.FindElement(By.Id("CardPanel"));
            // Получаем блок основных данных
            var mainBlock = cardPanel.FindElement(By.Name("$CardTypes_Block_NotificationRule"));
            var titleControl = mainBlock.FindElement(By.Name("$CardTypes_Controls_Title"));
            var titleEditor = titleControl.FindElement(By.Id("ValueEditor"));
            var notificationTypesControl = mainBlock.FindElement(By.Name("$CardTypes_Controls_NotificationTypes"));
            var notificationTypesEditor = notificationTypesControl.FindElement(By.Id("EntityTextBox"));
            var disallowControl = mainBlock.FindElement(By.Name("$CardTypes_Controls_Disallow"));
            var disallowCheckBox = disallowControl.FindElement(By.Id("CheckBox"));
            // Получаем блок с условиями
            var conditionsBlock = cardPanel.FindElement(By.Name("$CardTypes_Controls_Conditions"));
            var dataTable = conditionsBlock.FindElement(By.Id("DataTable"));
            cancellationToken.ThrowIfCancellationRequested();

            // композитное условие.
            bool _ =
                !string.IsNullOrWhiteSpace(titleEditor.Text) &&
                !string.IsNullOrWhiteSpace(notificationTypesEditor.Text) &&
                disallowCheckBox.Selected &&
                dataTable.FindElement(By.XPath("*[@ClassName='GridViewItem' and descendant::*[@AutomationId='ConditionTypeID:bdabb5a6-3bea-41dc-9e1b-c9d2cb7fd54a'] and descendant::*[@AutomationId='Description' and @Name='Тип документа: Входящий; ']]")) is not null;

            // Второй вариант: ждём пока пользователь выполнит сказанное:
            //await GuideHelper.WaitSuccessActionOrThrowAsync<GuideDiscardException>(async () =>
            //    !string.IsNullOrWhiteSpace(titleEditor.Text) &&
            //    !string.IsNullOrWhiteSpace(notificationTypesEditor.Text) &&
            //    disallowCheckBox.Selected &&
            //    dataTable.FindElement(By.XPath("*[@ClassName='GridViewItem' and descendant::*[@AutomationId='ConditionTypeID:bdabb5a6-3bea-41dc-9e1b-c9d2cb7fd54a'] and descendant::*[@AutomationId='Description' and @Name='Тип документа: Входящий; ']]")) is not null,
            //    cancellationToken: cancellationToken);
        }

        /// <summary>
        /// Ожидаем от пользователя настройки правила.
        /// </summary>
        /// <param name="myPreferencesWindow">Окно "Мои настройки".</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task SaveNotificationRuleAsync(IWebElement myPreferencesWindow, CancellationToken cancellationToken)
        {
            var ruleWindow = this.Provider.MainWindow.FindElement(By.XPath("*[@ClassName='Window' and @AutomationId='CardTableRowEditorDialog']"));
            // Получим название правила.
            var cardPanel = ruleWindow.FindElement(By.Id("CardPanel"));
            // Получаем блок основных данных
            var mainBlock = cardPanel.FindElement(By.Name("$CardTypes_Block_NotificationRule"));
            var titleControl = mainBlock.FindElement(By.Name("$CardTypes_Controls_Title"));
            var titleEditor = titleControl.FindElement(By.Id("ValueEditor"));
            cancellationToken.ThrowIfCancellationRequested();
            var title = titleEditor.Text;

            // Подсказка, что ожидается.
            await this.Provider.ShowTooltipAsync(
                "Настройте правило нажав на кнопку \"Сохранить\" в окне \"Правило уведомления\".\nПодтвердите выполнение настроек нажав кнопку \"Готово\".",
                ruleWindow,
                x => x
                    .UseInteractionColor()
                    .UseButtons(StandardButtons.Done)
                    .CloseAfterMaxInteractionDelay(1),
                cancellationToken);

            cardPanel = myPreferencesWindow.FindElement(By.Id("CardPanel"));
            // Получаем блок с условиями
            var conditionsBlock = cardPanel.FindElement(By.Id("Rules"));
            var dataTable = conditionsBlock.FindElement(By.Id("DataTable"));
            cancellationToken.ThrowIfCancellationRequested();

            // Проверяем выполнение с ожиданием.
            bool _ =
                dataTable.FindElement(By.XPath($"*[@ClassName='GridViewItem' and descendant::*[@AutomationId='Name' and @Name='{title}']]")) is not null;
        }

        /// <summary>
        /// Ожидаем от пользователя закрытия диалога "Мои настройки".
        /// </summary>
        /// <param name="window">Окно "Мои настройки".</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task ClosePreferencesAsync(IWebElement window, CancellationToken cancellationToken)
        {
            // Подсказка, что ожидается.
            await using (await this.Provider.ShowTooltipDisposableAsync(
                "Сохраните сделанные изменения в диалоге \"Мои настройки\", нажав на кнопку \"Сохранить\".",
                window, x => x.UseInteractionColor()))
            {
                // Проверяем выполнение с ожиданием.
                await this.Provider.WaitSuccessActionOrThrowAsync<GuideDiscardException>(async () =>
                    this.Provider.MainWindow.FindElements(By.XPath("*[@ClassName='Window' and @Name='Мои настройки']")).Count == 0,
                    cancellationToken: cancellationToken);
            }
        }

        #endregion
    }
}
