﻿using System;
using System.Threading.Tasks;
using System.Windows;
using UserGuide.Extensions;
using OpenQA.Selenium;

namespace UserGuide.Windows
{
    /// <summary>
    /// Обёртка над окном подсветки.
    /// </summary>
    public class Highlighter : IAsyncDisposable
    {
        #region Private fields

        /// <summary>
        /// Окна "подсветки", которыми управляет данный класс.
        /// </summary>
        private HighlighterWindow[] windows;

        #endregion
        
        /// <summary>
        /// Открыть окно для подсветки элемента.
        /// </summary>
        /// <param name="rectFuncs">Функции, возвращающие ограничивающие прямоугольники для элементов.</param>
        /// <returns>Задача, возвращающая удобную обёртку для взаимодействия с окном.</returns>
        public static async Task<Highlighter> ShowAsync(params Func<Rect>[] rectFuncs)
        {
            if (rectFuncs is null || rectFuncs.Length < 1)
            {
                return null;
            }
            var highlighter = new Highlighter();
            highlighter.windows = new HighlighterWindow[rectFuncs.Length];
            for (int i = 0; i < rectFuncs.Length; ++i)
            {
                highlighter.windows[i] = await HighlighterWindow.ShowAsync(rectFuncs[i]);
            }
            return highlighter;
        }

        /// <summary>
        /// Открыть окно для подсветки элемента.
        /// </summary>
        /// <param name="targetRects">Ограничивающие прямоугольники для элементов.</param>
        /// <returns>Задача, возвращающая удобную обёртку для взаимодействия с окном.</returns>
        public static async Task<Highlighter> ShowAsync(params Rect[] targetRects)
        {
            if (targetRects is null || targetRects.Length < 1)
            {
                return null;
            }
            var highlighter = new Highlighter();
            highlighter.windows = new HighlighterWindow[targetRects.Length];
            for (int i = 0; i < targetRects.Length; ++i)
            {
                highlighter.windows[i] = await HighlighterWindow.ShowAsync(targetRects[i]);
            }
            return highlighter;
        }

        /// <summary>
        /// Открыть окно для подсветки элементов.
        /// </summary>
        /// <param name="elements">Элементы.</param>
        /// <returns>Задача, возвращающая удобную обёртку для взаимодействия с окном.</returns>
        public static async Task<Highlighter> ShowAsync(params IWebElement[] elements)
        {
            if (elements is null || elements.Length < 1)
            {
                return null;
            }
            var highlighter = new Highlighter();
            highlighter.windows = new HighlighterWindow[elements.Length];
            for (int i = 0; i < elements.Length; ++i)
            {
                var l = i;
                highlighter.windows[i] = await HighlighterWindow.ShowAsync(() => elements[l].GetRect());
            }
            return highlighter;
        }

        /// <summary>
        /// Закрыть окно подсветки элементов.
        /// </summary>
        /// <returns>Задача.</returns>
        public async Task CloseAsync()
        {
            if (this.windows is null)
            {
                return;
            }

            foreach (var window in this.windows)
            {
                await window.CloseAsync();
            }

            this.windows = null;
        }

        /// <inheritdoc/>
        public async ValueTask DisposeAsync()
        {
            await this.CloseAsync();
        }
    }
}
