﻿using System;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Threading;

namespace UserGuide.Windows
{
    /// <summary>
    /// Interaction logic for HighlighterWindow.xaml
    /// </summary>
    public partial class HighlighterWindow : Window
    {
        #region Internals

        /// <summary>
        /// Класс для хранения состояния окна.
        /// </summary>
        class CapturedState
        {
            /// <summary>
            /// Прямоугольник, вокруг которого отображено окно.
            /// </summary>
            public Rect ElementRect { get; set; }

            /// <summary>
            /// Функция для получения (обновления) прямоугольника.
            /// </summary>
            public Func<Rect> ElementRectFunc { get; init; }
        }

        #endregion

        #region Private fields and consts

        /// <summary>
        /// Внешнее смещение рамки относительно ограничивающего прямоугольника.
        /// Рамку можно сделать как шире, так и уже.
        /// </summary>
        private const int Offset = 2;

        /// <summary>
        /// Таймер обновления данных о прямоугольнике
        /// и перемещения окна вслед за прямоугольником.
        /// </summary>
        private DispatcherTimer refreshTimer;

        /// <summary>
        /// Состояние окна.
        /// </summary>
        private CapturedState state;

        #endregion

        #region Ctor

        /// <summary>
        /// Конструктор по умолчанию.
        /// </summary>
        public HighlighterWindow()
        {
            this.InitializeComponent();
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Открыть окно для подсветки элемента.
        /// </summary>
        /// <param name="rectFunc">Функция для получения ограничивающего прямоугольника целевого элемента.</param>
        /// <returns>Задача, возвращающая удобную обёртку для взаимодействия с окном.</returns>
        public static async Task<HighlighterWindow> ShowAsync(Func<Rect> rectFunc)
        {
            return await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                var window = new HighlighterWindow
                {
                    state = new()
                    {
                        ElementRectFunc = rectFunc,
                        ElementRect = rectFunc(),
                    }
                };
                window.PrepareWindow();
                window.Show();
                return window;
            });
        }

        /// <summary>
        /// Открыть окно для подсветки элемента.
        /// </summary>
        /// <param name="targetRect">Ограничивающий прямоугольник элемента.</param>
        /// <returns>Задача, возвращающая удобную обёртку для взаимодействия с окном.</returns>
        public static async Task<HighlighterWindow> ShowAsync(Rect targetRect)
        {
            return await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                var window = new HighlighterWindow
                {
                    state = new()
                    {
                        ElementRectFunc = null,
                        ElementRect = targetRect,
                    }
                };
                window.PrepareWindow();
                window.Show();
                return window;
            });
        }

        /// <summary>
        /// Закрыть все окна подсветки элементов.
        /// </summary>
        /// <returns>Задача.</returns>
        public async Task CloseAsync()
        {
            await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                this.refreshTimer?.Stop();
                this.refreshTimer = null;
                this.Close();
            });
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Настроить окно для отображения.
        /// </summary>
        private void PrepareWindow()
        {
            if (this.state.ElementRectFunc is not null)
            {
                this.refreshTimer = new() { Interval = TimeSpan.FromSeconds(1) };
                this.refreshTimer.Tick += (o, e) =>
                {
                    try
                    {
                        if (this.state.ElementRectFunc is null)
                        {
                            return;
                        }
                        var rect = this.state.ElementRectFunc();
                        if (Rect.Equals(rect, this.state.ElementRect))
                        {
                            return;
                        }
                        this.state.ElementRect = rect;
                        this.SetPosition();
                    }
                    catch
                    {
                        // Игнорируем всё.
                    }
                };
            }
            this.Loaded += (o, e) =>
            {
                this.SetPosition();
                this.Border.Visibility = Visibility.Visible;
                this.refreshTimer?.Start();
            };
        }

        /// <summary>
        /// Установить окно в заданную позицию.
        /// </summary>
        private void SetPosition()
        {
            this.Left = this.state.ElementRect.Left - Offset;
            this.Top = this.state.ElementRect.Top - Offset;
            this.Width = this.state.ElementRect.Width + Offset * 2;
            this.Height = this.state.ElementRect.Height + Offset * 2;
        }

        #endregion
    }
}
