﻿using System;
using System.Globalization;
using System.Windows;
using System.Windows.Documents;
using Markdig;
using Markdig.Renderers;
using Markdig.Renderers.Wpf;
using Markdig.Syntax;
using Markdig.Wpf;


namespace UserGuide.Windows
{
    /// <summary>
    /// Помощник генерации <see cref="FlowDocument"/> из Markdown-а.
    /// </summary>
    public static class MarkdownHelper
    {
        #region Helper Classes

        private sealed class CustomWpfRenderer : WpfRenderer
        {
            protected override void LoadRenderers()
            {
                this.ObjectRenderers.AddIfNotAlready<CustomCodeBlockRenderer>();
                this.ObjectRenderers.AddIfNotAlready<CustomHeadingRenderer>();
                this.ObjectRenderers.AddIfNotAlready<CustomListRenderer>();
                this.ObjectRenderers.AddIfNotAlready<CustomParagraphRenderer>();
                base.LoadRenderers();
            }
        }

        private sealed class CustomParagraphRenderer : ParagraphRenderer
        {
            protected override void Write(WpfRenderer renderer, ParagraphBlock obj)
            {
                if (renderer == null) throw new ArgumentNullException(nameof(renderer));
                if (obj == null) throw new ArgumentNullException(nameof(obj));

                var paragraph = new Paragraph { Margin = new(0) };
                paragraph.SetResourceReference(FrameworkContentElement.StyleProperty, Styles.ParagraphStyleKey);

                renderer.Push(paragraph);
                renderer.WriteLeafInline(obj);
                renderer.Pop();
            }
        }

        public class CustomCodeBlockRenderer : CodeBlockRenderer
        {
            protected override void Write(WpfRenderer renderer, CodeBlock obj)
            {
                if (renderer == null) throw new ArgumentNullException(nameof(renderer));

                var paragraph = new Paragraph { Margin = new(0) };
                paragraph.SetResourceReference(FrameworkContentElement.StyleProperty, Styles.CodeBlockStyleKey);
                renderer.Push(paragraph);
                renderer.WriteLeafRawLines(obj);
                renderer.Pop();
            }
        }

        public class CustomHeadingRenderer : HeadingRenderer
        {
            protected override void Write(WpfRenderer renderer, HeadingBlock obj)
            {
                if (renderer == null) throw new ArgumentNullException(nameof(renderer));
                if (obj == null) throw new ArgumentNullException(nameof(obj));

                var paragraph = new Paragraph { Margin = new(0) };
                ComponentResourceKey styleKey = null;

                switch (obj.Level)
                {
                    case 1: styleKey = Styles.Heading1StyleKey; break;
                    case 2: styleKey = Styles.Heading2StyleKey; break;
                    case 3: styleKey = Styles.Heading3StyleKey; break;
                    case 4: styleKey = Styles.Heading4StyleKey; break;
                    case 5: styleKey = Styles.Heading5StyleKey; break;
                    case 6: styleKey = Styles.Heading6StyleKey; break;
                }

                if (styleKey != null)
                {
                    paragraph.SetResourceReference(FrameworkContentElement.StyleProperty, styleKey);
                }

                renderer.Push(paragraph);
                renderer.WriteLeafInline(obj);
                renderer.Pop();
            }
        }

        public class CustomListRenderer : ListRenderer
        {
            protected override void Write(WpfRenderer renderer, ListBlock listBlock)
            {
                if (renderer == null) throw new ArgumentNullException(nameof(renderer));
                if (listBlock == null) throw new ArgumentNullException(nameof(listBlock));

                var list = new List { Margin = new(0), Padding = new(25, 0, 0, 0) };

                if (listBlock.IsOrdered)
                {
                    list.MarkerStyle = TextMarkerStyle.Decimal;

                    if (listBlock.OrderedStart != null && listBlock.DefaultOrderedStart != listBlock.OrderedStart)
                    {
                        list.StartIndex = int.Parse(listBlock.OrderedStart, NumberFormatInfo.InvariantInfo);
                    }
                }
                else
                {
                    list.MarkerStyle = TextMarkerStyle.Disc;
                }

                renderer.Push(list);

                foreach (var item in listBlock)
                {
                    var listItemBlock = (ListItemBlock) item;
                    var listItem = new ListItem();
                    renderer.Push(listItem);
                    renderer.WriteChildren(listItemBlock);
                    renderer.Pop();
                }

                renderer.Pop();
            }
        }

        #endregion

        #region Static Methods

        /// <summary>
        /// Преобразовать текст в формате Markdown во <see cref="FlowDocument"/>.
        /// </summary>
        /// <param name="markdownText">Текст для преобразования.</param>
        /// <returns><see cref="FlowDocument"/></returns>
        public static FlowDocument ToFlowDocument(string markdownText) =>
            Markdig.Wpf.Markdown.ToFlowDocument(
                markdownText,
                new MarkdownPipelineBuilder()
                    .UseSupportedExtensions()
                    .UseEmojiAndSmiley()
                    .Build(),
                renderer: new CustomWpfRenderer());

        #endregion
    }
}
