﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using UserGuide.Extensions;
using OpenQA.Selenium;

namespace UserGuide.Windows
{
    /// <summary>
    /// Удобная обёртка над <see cref="TooltipWindow"/>.
    /// </summary>
    public class Tooltip : IAsyncDisposable
    {
        #region Ctor

        private Tooltip(TooltipWindow window) =>
            this.window = window;

        #endregion

        #region Private fields

        /// <summary>
        /// Окно с подсказкой.
        /// </summary>
        private readonly TooltipWindow window;

        #endregion

        #region Public methods

        /// <summary>
        /// Показать окно с заданным текстом в заданном положении относительно указанной области.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="rectFunc">Функция, возвращающая ограничивающий прямоугольник для заданного элемента.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <returns>Задача с обёрткой над окном подсказки.</returns>
        public static async Task<Tooltip> ShowAsync(
            string text,
            Func<Rect> rectFunc,
            TooltipOptions options)
        {
            return new Tooltip(await TooltipWindow.ShowAsync(text, rectFunc, options));
        }

        /// <summary>
        /// Показать окно с заданным текстом в заданном положении относительно указанной области.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="element">Целевой элемент.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <returns>Задача с обёрткой над окном подсказки.</returns>
        public static async Task<Tooltip> ShowAsync(
            string text,
            IWebElement element,
            TooltipOptions options)
        {
            return new Tooltip(await TooltipWindow.ShowAsync(text, element.GetRect, options));
        }

        /// <summary>
        /// Показать окно с заданным текстом в заданном положении относительно указанной области.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="targetRect">Целевой прямоугольник.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <returns>Задача с обёрткой над окном подсказки.</returns>
        public static async Task<Tooltip> ShowAsync(
            string text,
            Rect targetRect,
            TooltipOptions options)
        {
            return new Tooltip(await TooltipWindow.ShowAsync(text, targetRect, options));
        }

        /// <summary>
        /// Подождать закрытия окна.
        /// </summary>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Задача с результатом закрытия окна.</returns>
        public async Task<int> WaitUntilWindowClosedAsync(CancellationToken cancellationToken = default)
        {
            try
            {
                return await this.window.WindowClosedTask.WaitAsync(cancellationToken);
            }
            catch (OperationCanceledException)
            {
                await this.CloseAsync();
                throw;
            }
        }

        /// <summary>
        /// Закрыть окно немедленно.
        /// </summary>
        /// <returns>Асинхронная задача.</returns>
        public ValueTask CloseAsync() =>
            new(this.window.CloseAsync());

        /// <inheritdoc/>
        public ValueTask DisposeAsync() =>
            this.CloseAsync();

        #endregion
    }
}
