﻿using System;
using System.Windows.Controls;
using System.Windows.Media;

namespace UserGuide.Windows
{
    /// <summary>
    /// Помощник для построения опций отображения окна подсказки.
    /// </summary>
    public class TooltipOptionsBuilder
    {
        #region Private fields

        /// <summary>
        /// Поставщик настроек.
        /// </summary>
        private readonly GuideProvider provider;

        /// <summary>
        /// Опции окна с подсказкой.
        /// </summary>
        private TooltipOptions options;

        #endregion

        /// <summary>
        /// Создаёт построитель с экземпляром свойств по умолчанию.
        /// </summary>
        /// <param name="provider">Провайдер.</param>
        public TooltipOptionsBuilder(GuideProvider provider)
        {
            this.provider = provider;
            this.options = new()
            {
                Mode = Mode.AutomationTooltip,
                Placement = Placement.Bottom,
                ArrowPlacement = ArrowPlacement.Center,
                ButtonsOrientation = Orientation.Horizontal,
                CloseOnClick = false,
            };
        }

        #region Public methods

        /// <summary>
        /// Построить опции открытия окна.
        /// </summary>
        /// <returns>Опции открытия окна, согласно произведённым настройкам.</returns>
        public TooltipOptions Build() =>
            this.options;

        /// <summary>
        /// Отобразить окно в режиме подсказки автоматизации.
        /// <para>
        /// Внимание! Этот режим установлен по умолчанию!
        /// </para>
        /// </summary>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder UseAutomationColor()
        {
            this.options = this.options with { Mode = Mode.AutomationTooltip };
            return this;
        }

        /// <summary>
        /// Отобразить в режиме пользовательского диалога.
        /// </summary>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder UseInteractionColor()
        {
            this.options = this.options with { Mode = Mode.UserDialog };
            return this;
        }

        /// <summary>
        /// Отобразить в настраиваемом режиме с заданным цветом диалога.
        /// </summary>
        /// <param name="color">Цвет фона диалога.</param>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder UseCustomColor(Color color)
        {
            this.options = this.options with
            { 
                Mode = Mode.Custom,
                CustomColor = color,
            };
            return this;
        }

        /// <summary>
        /// Отметить, что текст диалога задан в формате Markdown.
        /// </summary>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder UseMarkdown()
        {
            this.options = this.options with { MessageInMarkdown = true };
            return this;
        }

        /// <summary>
        /// Задать положение диалога относительно целевого элемента.
        /// <para>
        /// Внимание! По умолчанию установлено <c>placement</c> = <see cref="Placement.Bottom"/> и <c>arrowPlacement</c> = <see cref="ArrowPlacement.Center"/>.
        /// </para>
        /// </summary>
        /// <param name="placement">Главное положение.</param>
        /// <param name="arrowPlacement">Детализирующее положение для треугольников (стрелок).</param>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder Place(
            Placement placement,
            ArrowPlacement arrowPlacement = ArrowPlacement.Center)
        {
            this.options = this.options with { Placement = placement, ArrowPlacement = arrowPlacement };
            return this;
        }

        /// <summary>
        /// Установить явный режим закрытия окна при клике на нём.
        /// <para>
        /// Внимание! По умолчанию установлен режим <c>closeOnClick</c> = <c>false</c>.
        /// </para>
        /// </summary>
        /// <param name="value"><c>true</c> - закрывать окно при клике на нём, <c>false</c> - не закрывать окно при клике на нём.</param>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder CloseOnClick(bool value = true)
        {
            this.options = this.options with { CloseOnClick = value };
            return this;
        }

        /// <summary>
        /// Установить время закрытия окна.
        /// </summary>
        /// <param name="multiplier">Коэффициент относительно базовой задержки.</param>
        /// <returns>Построитель.</returns>
        /// <exception cref="ArgumentOutOfRangeException">Передан отрицательный или нулевой коэффициент.</exception>
        public TooltipOptionsBuilder CloseAfter(double multiplier)
        {
            if (multiplier <= 0)
            {
                throw new ArgumentOutOfRangeException(nameof(multiplier), $"Передано недопустимое значение \"{multiplier}\". Значение должно быть больше нуля.");
            }
            this.options = this.options with { ShowTime = this.provider.DialogDelay * multiplier };
            return this;
        }

        /// <summary>
        /// Установить время закрытия окна.
        /// </summary>
        /// <param name="multiplier">Коэффициент относительно базовой задержки реакции пользователя.</param>
        /// <returns>Построитель.</returns>
        /// <exception cref="ArgumentOutOfRangeException">Передан отрицательный или нулевой коэффициент.</exception>
        public TooltipOptionsBuilder CloseAfterMaxInteractionDelay(double multiplier)
        {
            if (multiplier <= 0)
            {
                throw new ArgumentOutOfRangeException(nameof(multiplier), $"Передано недопустимое значение \"{multiplier}\". Значение должно быть больше нуля.");
            }
            this.options = this.options with { ShowTime = this.provider.MaxInteractionDelay * multiplier };
            return this;
        }

        /// <summary>
        /// Задаёт используемые в окне кнопки и их расположение.
        /// </summary>
        /// <param name="descriptors">Описание кнопок.</param>
        /// <param name="orientation">Положение кнопок.</param>
        /// <returns>Построитель.</returns>
        /// <exception cref="ArgumentException">Не задано ни одной кнопки.</exception>
        public TooltipOptionsBuilder UseButtons(
            ButtonDescriptor[] descriptors,
            Orientation orientation)
        {
            if (descriptors is null || descriptors.Length < 1)
            {
                throw new ArgumentException("Должна быть задана хотя бы одна кнопка.", nameof(descriptors));
            }
            this.options = this.options with { ButtonDescriptors = descriptors, ButtonsOrientation = orientation };
            return this;
        }

        /// <summary>
        /// Задаёт используемые в окне кнопки расположенные по горизонтали.
        /// </summary>
        /// <param name="descriptors">Описание кнопок.</param>
        /// <returns>Построитель.</returns>
        /// <exception cref="ArgumentException">Не задано ни одной кнопки.</exception>
        public TooltipOptionsBuilder UseButtons(params ButtonDescriptor[] descriptors)
        {
            if (descriptors is null || descriptors.Length < 1)
            {
                throw new ArgumentException("Должна быть задана хотя бы одна кнопка.", nameof(descriptors));
            }
            this.options = this.options with { ButtonDescriptors = descriptors, ButtonsOrientation = Orientation.Horizontal };
            return this;
        }

        /// <summary>
        /// Задаёт расположение кнопок окна (по вертикали).
        /// </summary>
        /// <param name="orientation">Положение кнопок.</param>
        /// <returns>Построитель.</returns>
        public TooltipOptionsBuilder UseButtonsOrientation(
            Orientation orientation = Orientation.Vertical)
        {
            this.options = this.options with { ButtonsOrientation = orientation };
            return this;
        }
        
        #endregion
    }
}
