﻿using System;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;
using UserGuide.Extensions;
using Emoji.Wpf;

namespace UserGuide.Windows
{
    /// <summary>
    /// Interaction logic for TooltipWindow.xaml
    /// </summary>
    public partial class TooltipWindow : Window
    {
        #region Internals

        /// <summary>
        /// Описывает состояние окна.
        /// </summary>
        private sealed class CapturedState
        {
            /// <summary>
            /// Основное местоположение окна относительно прямоугольника.
            /// </summary>
            public Placement Placement { get; init; }

            /// <summary>
            /// Уточняющее местоположение окна относительно прямоугольника.
            /// </summary>
            public ArrowPlacement SubPlacement { get; init; }

            /// <summary>
            /// Последнее запомненное положение элемента (прямоугольник).
            /// </summary>
            public Rect ElementRect { get; set; }

            /// <summary>
            /// Функция получения положения ограничивающего прямоугольника.
            /// </summary>
            public Func<Rect> ElementRectFunc { get; init; }
        }

        /// <summary>
        /// Режим коррекции положения и размеров окна.
        /// </summary>
        private enum CorrectionMode
        {
            /// <summary>
            /// Коррекция не производилась.
            /// </summary>
            None,

            /// <summary>
            /// Скорректировано положение и ширина окна.
            /// </summary>
            Width,

            /// <summary>
            /// Скорректировано положение окна.
            /// </summary>
            Offset,
        }

        #endregion

        #region Private fields and consts

        /// <summary>
        /// Фон подсказки, когда действия выполняются только автоматизацией.
        /// </summary>
        private static readonly Brush AutomationTooltipBrush = Brushes.LightBlue;

        /// <summary>
        /// Фон подсказки, когда действия должен выполнить пользователь.
        /// </summary>
        private static readonly Brush UserActionsTooltipBrush = Freeze(new SolidColorBrush(Color.FromArgb(255, 255, 209, 198)));

        /// <summary>
        /// Высота треугольника.
        /// </summary>
        private const int TriangleHeight = 25;

        /// <summary>
        /// Минимальная ширина треугольника.
        /// </summary>
        private const int MinTriangleWidth = 10;

        /// <summary>
        /// Максимальная ширина треугольника.
        /// </summary>
        private const int MaxTriangleWidth = 2 * TriangleHeight;

        /// <summary>
        /// Размер скругления углов рамки.
        /// </summary>
        private const int BorderCorner = 10;

        /// <summary>
        /// Код закрытия диалога по умолчанию.
        /// </summary>
        private const int DefaultDialogResult = -1;

        /// <summary>
        /// Состояние диалога.
        /// </summary>
        private readonly CapturedState state;

        /// <summary>
        /// Кисть для фона диалога.
        /// </summary>
        private Brush backgroundBrush;
        
        /// <summary>
        /// Источник задачи с результатом закрытия окна.
        /// </summary>
        private readonly TaskCompletionSource<int> windowClosedSource = new();

        /// <summary>
        /// Таймер закрытия окна, если нужно отображать окно в течение заданного времени.
        /// </summary>
        private DispatcherTimer closeTimer;

        /// <summary>
        /// Таймер обновления окна, если целевой прямоугольник не статичен.
        /// </summary>
        private DispatcherTimer refreshTimer;

        /// <summary>
        /// Признак, что нужно закрывать окно при клике мыши по любой его области кроме кнопок.
        /// </summary>
        private bool closeOnClick;

        #endregion

        #region Public fields

        /// <summary>
        /// Использовать алгоритм коррекции положения окна.
        /// </summary>
        public static bool UseWindowCorrection = true;

        #endregion

        #region Ctor

        /// <summary>
        /// Создаёт новое окно.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <param name="state">Состояние окна.</param>
        private TooltipWindow(
            string text,
            TooltipOptions options,
            CapturedState state)
        {
            this.InitializeComponent();
            this.state = state;
            this.PrepareWindow(text, options);
        }

        #endregion

        #region Public properties

        /// <summary>
        /// Результат закрытия окна (актуально при наличии нескольких кнопок).
        /// По умолчанию <see cref="DefaultDialogResult"/>.
        /// </summary>
        public int ResultCode { get; set; } = DefaultDialogResult;

        /// <summary>
        /// Задача закрытия окна (по действиям пользователя или таймеру).
        /// </summary>
        public Task<int> WindowClosedTask => this.windowClosedSource.Task;

        #endregion

        #region Public methods

        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="targetRect">Целевой прямоугольник.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <returns>Задача, возвращающая окно.</returns>
        public static async Task<TooltipWindow> ShowAsync(
            string text,
            Rect targetRect,
            TooltipOptions options)
        {
            return await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                var window = new TooltipWindow(
                    text,
                    options,
                    new CapturedState()
                    {
                        Placement = options.Placement,
                        SubPlacement = options.ArrowPlacement,
                        ElementRect = targetRect,
                        ElementRectFunc = null,
                    });

                window.Show();
                return window;
            });
        }

        /// <summary>
        /// Показать окно с заданным текстом, в заданном положении относительно указанной области.
        /// </summary>
        /// <param name="text">Отображаемый текст.</param>
        /// <param name="rectFunc">Функция, возвращающая ограничивающий прямоугольник для целевого элемента.</param>
        /// <param name="options">Настройки диалога.</param>
        /// <returns>Задача, возвращающая окно.</returns>
        public static async Task<TooltipWindow> ShowAsync(
            string text,
            Func<Rect> rectFunc,
            TooltipOptions options)
        {
            return await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                var window = new TooltipWindow(
                    text,
                    options,
                    new CapturedState()
                    {
                        Placement = options.Placement,
                        SubPlacement = options.ArrowPlacement,
                        ElementRectFunc = rectFunc,
                        ElementRect = rectFunc(),
                    });
                
                window.Show();
                return window;
            });
        }

        /// <summary>
        /// Помощник закрытия окна.
        /// </summary>
        /// <returns>Задача закрытия окна.</returns>
        public Task CloseAsync()
        {
            if (this.Dispatcher.CheckAccess())
            {
                this.Close();
                return Task.CompletedTask;
            }

            return this.Dispatcher.InvokeAsync(this.Close).Task;
        }

        #endregion

        #region Protected methods

        /// <inheritdoc/>
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            if (this.closeOnClick)
            {
                this.Close();
            }
        }

        #endregion

        #region Private methods

        private void PrepareWindow(
            string text,
            TooltipOptions options)
        {
            if (this.state.ElementRectFunc is not null)
            {
                this.refreshTimer = new() { Interval = TimeSpan.FromSeconds(1) };
                this.refreshTimer.Tick += (o, e) =>
                {
                    try
                    {
                        if (this.state.ElementRectFunc is null)
                        {
                            return;
                        }
                        var rect = this.state.ElementRectFunc();
                        if (Rect.Equals(rect, this.state.ElementRect))
                        {
                            return;
                        }
                        this.state.ElementRect = rect;
                        this.SetWindowPosition();
                    }
                    catch
                    {
                        // Игнорируем всё.
                    }
                };
            }

            this.Border.Background = this.GetBrush(options);
            if (options.MessageInMarkdown)
            {
                var flowDoc = MarkdownHelper.ToFlowDocument(text);
                flowDoc.SubstituteGlyphs();
                flowDoc.PagePadding = new(0);

                this.RichMessage.Document = flowDoc;
            }
            else
            {
                this.Message.Text = text;
            }

            this.Loaded += (o, e) =>
            {
                this.SetWindowPosition();
                this.closeTimer?.Start();
                this.refreshTimer?.Start();
            };

            this.Closed += this.OnClosed;

            if (options.ShowTime is not null && options.ShowTime.Value.TotalMilliseconds > 0)
            {
                this.closeTimer = new() { Interval = options.ShowTime.Value };
                this.closeTimer.Tick += (o, e) =>
                {
                    this.Close();
                };
            }
            if (options.ButtonDescriptors is { Length: > 0 })
            {
                this.ButtonsPanel.Margin = new Thickness(0, 10, 0, 0);
                DependencyProperty placementProperty = Grid.ColumnProperty;
                if (options.ButtonsOrientation == Orientation.Horizontal)
                {
                    for (int i = 0; i < options.ButtonDescriptors.Length; ++i)
                    {
                        this.ButtonsPanel.ColumnDefinitions.Add(new ColumnDefinition());
                    }
                }
                else
                {
                    placementProperty = Grid.RowProperty;
                    for (int i = 0; i < options.ButtonDescriptors.Length; ++i)
                    {
                        this.ButtonsPanel.RowDefinitions.Add(new RowDefinition());
                    }
                }
                for (int i = 0; i < options.ButtonDescriptors.Length; ++i)
                {
                    var buttonDescriptor = options.ButtonDescriptors[i];
                    var button = new Button { Content = buttonDescriptor.Text };
                    button.Click += (o, e) =>
                    {
                        this.ResultCode = buttonDescriptor.Code;
                        this.Close();
                    };
                    button.SetValue(placementProperty, i);
                    button.Margin = new Thickness(1);
                    button.Style = (Style) this.FindResource("CustomButtonStyle");
                    this.ButtonsPanel.Children.Add(button);
                }
            }
            if (options.CloseTask is not null)
            {
                options.CloseTask.ContinueWith(_ => this.CloseAsync(), TaskContinuationOptions.ExecuteSynchronously);
            }
            this.closeOnClick = options.CloseOnClick;
        }

        private void OnClosed(object sender, EventArgs e)
        {
            if (this.closeTimer is not null)
            {
                this.closeTimer.Stop();
                this.closeTimer = null;
            }
            if (this.refreshTimer is not null)
            {
                this.refreshTimer.Stop();
                this.refreshTimer = null;
            }

            this.windowClosedSource.SetResult(this.ResultCode);
        }

        private Brush GetBrush(TooltipOptions options) =>
            this.backgroundBrush ??= options.Mode switch
            {
                Mode.UserDialog => UserActionsTooltipBrush,
                Mode.AutomationTooltip => AutomationTooltipBrush,
                Mode.Custom => options.CustomColor is null
                    ? throw new InvalidOperationException(
                        "Задан настраиваемый режим отображения окна с подсказкой, но не указан цвет фона.")
                    : Freeze(new SolidColorBrush(options.CustomColor.Value)),
                _ => throw new InvalidOperationException("Задан неизвестный режим отображения диалога с подсказкой.")
            };

        /// <summary>
        /// Заморозить кисть.
        /// </summary>
        /// <param name="brush">Кисть.</param>
        /// <returns>Замороженная кисть.</returns>
        private static Brush Freeze(Brush brush)
        {
            brush.Freeze();
            return brush;
        }

        private void SetWindowPosition() =>
            this.SetWindowPosition(this.state.Placement, this.state.SubPlacement, this.state.ElementRect);

        private void SetWindowPosition(Placement placement, ArrowPlacement subPlacement, Rect targetRect)
        {
            // вначале отображаем треугольник, т.к. от этого меняется размер окна!
            this.ShowTriangle(placement, subPlacement);

            if (UseWindowCorrection)
            {
                this.MeasureWindowRect(targetRect, placement, subPlacement, out var windowRect, out var mode);

                // если меняем ширину окна, то нужно снова пересчитать его положение
                // т.к. может измениться высота...
                if (mode == CorrectionMode.Width)
                {
                    this.MaxWidth = windowRect.Width;
                    this.UpdateLayout();
                    var p = this.GetWindowPoint(new Size(this.ActualWidth, this.ActualHeight), targetRect, placement, subPlacement);
                    this.Left = p.X;
                    this.Top = p.Y;
                }
                else
                {
                    this.Left = windowRect.Left;
                    this.Top = windowRect.Top;
                }
                if (mode != CorrectionMode.None)
                {
                    this.CoerceTriangle(placement, subPlacement, targetRect);
                }
                return;
            }
            // старый алгоритм.
            var point = this.GetWindowPoint(new Size(this.ActualWidth, this.ActualHeight), targetRect, placement, subPlacement);
            this.Left = point.X;
            this.Top = point.Y;
        }

        private void ShowTriangle(Placement placement, ArrowPlacement subPlacement)
        {
            double width;
            switch (placement)
            {
                case Placement.Left:
                    // Right triangle
                    width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualHeight * 0.3));
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(0, 0),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(TriangleHeight, width * 0.5),
                                    new Point(0, width),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(-1, BorderCorner, 0, BorderCorner);
                    this.Triangle.VerticalAlignment = GetVerticalAlignment(subPlacement);
                    Grid.SetRow(this.Triangle, 1);
                    Grid.SetColumn(this.Triangle, 2);
                    break;
                case Placement.Right:
                    // Left triangle
                    width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualHeight * 0.3));
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(TriangleHeight, 0),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(0, width * 0.5),
                                    new Point(TriangleHeight, width),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(0, BorderCorner, -1, BorderCorner);
                    this.Triangle.VerticalAlignment = GetVerticalAlignment(subPlacement);
                    Grid.SetRow(this.Triangle, 1);
                    Grid.SetColumn(this.Triangle, 0);
                    break;
                case Placement.Top:
                    // Bottom triangle
                    width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualWidth * 0.3));
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(0, 0),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(width * 0.5, TriangleHeight),
                                    new Point(width, 0),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(BorderCorner, -1, BorderCorner, 0);
                    this.Triangle.HorizontalAlignment = GetHorizontalAlignment(subPlacement);
                    Grid.SetRow(this.Triangle, 2);
                    Grid.SetColumn(this.Triangle, 1);
                    break;
                case Placement.Bottom:
                    // Top triangle
                    width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualWidth * 0.3));
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(0, TriangleHeight),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(width * 0.5, 0),
                                    new Point(width, TriangleHeight),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(BorderCorner, 0, BorderCorner, -1);
                    this.Triangle.HorizontalAlignment = GetHorizontalAlignment(subPlacement);
                    Grid.SetRow(this.Triangle, 0);
                    Grid.SetColumn(this.Triangle, 1);
                    break;
                case Placement.Center:
                default:
                    return;
            }
            
            this.Triangle.Fill = this.Border.Background;
            this.Triangle.Visibility = Visibility.Visible;
            this.UpdateLayout();
        }

        private void CoerceTriangle(Placement placement, ArrowPlacement subPlacement, Rect targetRect)
        {
            double width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualWidth * 0.3));
            // Учтём отступы.
            double offsetLeft = BorderCorner, offsetRight = BorderCorner;
            switch (subPlacement)
            {
                case ArrowPlacement.Center:
                    {
                        double diff = Math.Max(0, BorderCorner - (targetRect.X - this.Left));
                        width = Math.Min(width, targetRect.Width - diff * 0.5);
                        offsetLeft = targetRect.X - this.Left + (targetRect.Width - width) * 0.5;
                        offsetRight = offsetLeft + width;
                    }
                    break;
                case ArrowPlacement.Left:
                    {
                        double diff = Math.Max(0, BorderCorner - (targetRect.X - this.Left));
                        width = Math.Min(width, (targetRect.Width - diff) * 0.5);
                        offsetLeft = diff > 0 ? BorderCorner : targetRect.X - this.Left;
                        offsetRight = offsetLeft + width;
                    }
                    break;
                case ArrowPlacement.Right:
                    {
                        double diff = Math.Max(0, BorderCorner - (this.Left + this.ActualWidth - targetRect.Right));
                        width = Math.Min(width, (targetRect.Width - diff) * 0.5);
                        offsetRight = diff > 0 ? BorderCorner : this.Left + this.ActualWidth - targetRect.Right;
                        offsetLeft = this.ActualWidth - offsetRight - width;
                    }
                    break;
            }

            switch (placement)
            {
                case Placement.Top:
                    // Bottom triangle
                    this.Triangle.HorizontalAlignment = HorizontalAlignment.Stretch;
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(0, 0),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(width * 0.5, TriangleHeight),
                                    new Point(width, 0),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(offsetLeft, -1, offsetRight, 0);
                    break;
                case Placement.Bottom:
                    // Top triangle
                    this.Triangle.HorizontalAlignment = HorizontalAlignment.Stretch;
                    width = Math.Max(MinTriangleWidth, Math.Min(MaxTriangleWidth, this.ActualWidth * 0.3));
                    this.Triangle.Data = new PathGeometry(new[]
                    {
                        new PathFigure(
                            new Point(0, TriangleHeight),
                            new[]
                            {
                                new PolyLineSegment(new[]
                                {
                                    new Point(width * 0.5, 0),
                                    new Point(width, TriangleHeight),
                                }, false)
                            },
                            true)
                    });
                    this.Triangle.Margin = new Thickness(offsetLeft, 0, offsetRight, -1);
                    break;
            }
        }

        private void MeasureWindowRect(
            Rect targetRect,
            Placement placement,
            ArrowPlacement subPlacement,
            out Rect windowRect,
            out CorrectionMode mode)
        {
            // 1. Расчёт ожидаемого положения и размеров окна.
            var windowSize = new Size(this.ActualWidth, this.ActualHeight);
            var windowPoint = this.GetWindowPoint(windowSize, targetRect, placement, subPlacement);

            // 2. Проверка необходимости корректировки           
            // Получаем рабочий стол, на котором отображается элемент.
            double cx = targetRect.X + targetRect.Width * 0.5;
            double cy = targetRect.Y + targetRect.Height * 0.5;
            // Конвертируем обратно в пиксели.
            cx = SystemParametersExtensions.GetPixels(cx);
            cy = SystemParametersExtensions.GetPixels(cy);
            // Получаем экран.
            var screen = System.Windows.Forms.Screen.FromPoint(
                new System.Drawing.Point((int) cx, (int) cy));

            // Переводим границы с учётом Dpi в дюймы.
            var screenRect = System.Drawing.Rectangle.FromLTRB(
                (int) SystemParametersExtensions.GetInches(screen.Bounds.Left),
                (int) SystemParametersExtensions.GetInches(screen.Bounds.Top),
                (int) SystemParametersExtensions.GetInches(screen.Bounds.Right),
                (int) SystemParametersExtensions.GetInches(screen.Bounds.Bottom));
            // Если рассчитанный нами прямоугольник полностью внутри экрана
            // нам повезло и корректировка не требуется
            if (screenRect.Contains(new System.Drawing.Rectangle(
                (int) windowPoint.X,
                (int) windowPoint.Y,
                (int) windowSize.Width,
                (int) windowSize.Height)))
            {
                windowRect = new Rect(windowPoint, windowSize);
                mode = CorrectionMode.None;
                return;
            }
            // Нам не повезло, есть пересечения.
            // Пересечения по вертикали связаны с количеством текста (и мы не будем их здесь обрабатывать).
            // Определим пересечения по горизонтали.
            // Если ширина прямоугольника больше ширины экрана, то мы его ограничим.
            if (screenRect.Width < windowSize.Width)
            {
                windowRect = new Rect(screenRect.X, windowPoint.Y, screenRect.Width, windowSize.Height);
                mode = CorrectionMode.Width;
                return;
            }
            // Теперь определим с какой стороны мы пересекли экран.
            mode = CorrectionMode.Offset;
            if (screen.Bounds.X > windowPoint.X)
            {
                // Пересекли слева.
                windowRect = new Rect(screenRect.X, windowPoint.Y, windowSize.Width, windowSize.Height);
            }
            else
            {
                windowRect = new Rect(screenRect.Right - windowSize.Width, windowPoint.Y, windowSize.Width, windowSize.Height);
            }
        }

        private Point GetWindowPoint(
            Size windowSize,
            Rect targetRect,
            Placement placement,
            ArrowPlacement subPlacement)
        {
            double halfElementWidth = targetRect.Width * 0.5,
            halfElementHeight = targetRect.Height * 0.5;
            double cx = targetRect.X + halfElementWidth,
                cy = targetRect.Y + halfElementHeight;

            double x, y;
            double windowWidth = windowSize.Width,
                windowHeight = windowSize.Height;
            double halfWindowWidth = windowWidth * 0.5,
                halfWindowHeight = windowHeight * 0.5;
            switch (placement)
            {
                case Placement.Left:
                    x = cx - halfElementWidth - halfWindowWidth * 2.0;
                    y = GetVerticalPosition(subPlacement, cy, halfElementHeight, halfWindowHeight);
                    break;

                case Placement.Right:
                    x = cx + halfElementWidth;
                    y = GetVerticalPosition(subPlacement, cy, halfElementHeight, halfWindowHeight);
                    break;

                case Placement.Top:
                    x = GetHorizontalPosition(subPlacement, cx, halfElementWidth, halfWindowWidth);
                    y = cy - halfElementHeight - halfWindowHeight * 2.0;
                    break;

                case Placement.Bottom:
                    x = GetHorizontalPosition(subPlacement, cx, halfElementWidth, halfWindowWidth);
                    y = cy + halfElementHeight;
                    break;

                case Placement.Center:
                default:
                    x = cx - halfWindowWidth;
                    y = cy - halfWindowHeight;
                    break;
            }
            return new Point(x, y);
        }

        private static double GetVerticalPosition(ArrowPlacement placement, double cy, double halfElementHeight, double halfWindowHeight) =>
            placement switch
            {
                ArrowPlacement.Top => cy - halfElementHeight,
                ArrowPlacement.Bottom => cy + halfElementHeight - halfWindowHeight * 2,
                _ => cy - halfWindowHeight,
            };

        private static VerticalAlignment GetVerticalAlignment(ArrowPlacement placement) =>
            placement switch
            {
                ArrowPlacement.Top => VerticalAlignment.Top,
                ArrowPlacement.Center => VerticalAlignment.Center,
                ArrowPlacement.Bottom => VerticalAlignment.Bottom,
                _ => VerticalAlignment.Stretch,
            };

        private static double GetHorizontalPosition(ArrowPlacement placement, double cx, double halfElementWidth, double halfWindowWidth) =>
            placement switch
            {
                ArrowPlacement.Left => cx - halfElementWidth,
                ArrowPlacement.Right => cx + halfElementWidth - halfWindowWidth * 2,
                _ => cx - halfWindowWidth,
            };

        private static HorizontalAlignment GetHorizontalAlignment(ArrowPlacement placement) =>
            placement switch
            {
                ArrowPlacement.Left => HorizontalAlignment.Left,
                ArrowPlacement.Center => HorizontalAlignment.Center,
                ArrowPlacement.Right => HorizontalAlignment.Right,
                _ => HorizontalAlignment.Stretch,
            };

        #endregion
    }
}
