﻿using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;

namespace UserGuide.Windows
{
    /// <summary>
    /// Interaction logic for TrainingControlWindow.xaml
    /// </summary>
    public partial class TrainingControlWindow : Window
    {
        #region Private fields

        /// <summary>
        /// Источник объектов для отмены асинхронных операций.
        /// </summary>
        private CancellationTokenSource cancellationTokenSource;

        /// <summary>
        /// Состояние окна (какая кнопка нажата пользователем).
        /// </summary>
        private TrainingControlState state;

        #endregion

        #region Ctor

        /// <summary>
        /// Конструктор по умолчанию.
        /// </summary>
        public TrainingControlWindow()
        {
            this.InitializeComponent();

            this.ResetState();
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Открыть окно в заданном положении.
        /// </summary>
        /// <param name="targetRect">Целевой прямоугольник.</param>
        /// <param name="placement">Общее местоположение.</param>
        /// <param name="subPlacement">Более точная настройка.</param>
        public static async Task<TrainingControlWindow>
            ShowAsync(
                Rect targetRect,
                Placement placement = Placement.Bottom,
                ArrowPlacement subPlacement = ArrowPlacement.Center
            )
        {
            return await DispatcherHelper.Dispatcher.InvokeAsync(() =>
            {
                var window = new TrainingControlWindow();
                window.Loaded += (o, e) =>
                {
                    window.SetPosition(targetRect, placement, subPlacement);
                };
                window.Show();
                return window;
            });
        }

        /// <summary>
        /// Закрыть окно.
        /// </summary>
        /// <returns>Задача.</returns>
        public Task CloseAsync()
        {
            if (this.Dispatcher.CheckAccess())
            {
                this.Close();
                if (this.cancellationTokenSource is not null)
                {
                    this.cancellationTokenSource.Dispose();
                }
                return Task.CompletedTask;
            }

            return this.Dispatcher.InvokeAsync(() =>
            {
                this.Close();
                if (this.cancellationTokenSource is not null)
                {
                    this.cancellationTokenSource.Dispose();
                }
            }).Task;
        }

        /// <summary>
        /// Объект для отмены асинхронной операции.
        /// </summary>
        public CancellationToken CancellationToken =>
            this.cancellationTokenSource.Token;

        /// <summary>
        /// Нажатая кнопка.
        /// </summary>
        public TrainingControlState State =>
            this.state;

        /// <summary>
        /// Сбросить состояние.
        /// </summary>
        public void ResetState()
        {
            if (this.cancellationTokenSource != null)
            {
                this.cancellationTokenSource.Dispose();
            }
            this.cancellationTokenSource = new();
            this.state = TrainingControlState.None;
        }

        /// <summary>
        /// Скрыть или показать кнопку "Меню".
        /// </summary>
        /// <param name="value">Флаг <c>true</c> - скрыть, <c>false</c> - показать.</param>
        public void HideMenuButton(bool value = true)
        {
            if (this.Dispatcher.CheckAccess())
            {
                this.MenuButton.Visibility = value ? Visibility.Collapsed : Visibility.Visible;
                return;
            }

            this.Dispatcher.Invoke(() => this.MenuButton.Visibility = value ? Visibility.Collapsed : Visibility.Visible);
        }

        #endregion

        #region Protected methods

        /// <inheritdoc/>
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonDown(e);

            this.DragMove();
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Установить окно в заданную позицию, согласно настройкам.
        /// </summary>
        /// <param name="targetRect">Целевой прямоугольник.</param>
        /// <param name="placement">Основное положение.</param>
        /// <param name="subPlacement">Уточняющее положение.</param>
        private void SetPosition(
            Rect targetRect,
            Placement placement,
            ArrowPlacement subPlacement)
        {
            double halfElementWidth = targetRect.Width * 0.5,
            halfElementHeight = targetRect.Height * 0.5;
            double cx = targetRect.X + halfElementWidth,
                cy = targetRect.Y + halfElementHeight;
            double halfWindowWidth = this.ActualWidth * 0.5,
                halfWindowHeight = this.ActualHeight * 0.5;

            switch (placement)
            {
                case Placement.Left:
                    this.Left = cx - halfElementWidth;
                    this.Top = GetVerticalPosition(subPlacement, cy, halfElementHeight, halfWindowHeight);
                    break;
                case Placement.Right:
                    this.Left = cx + halfElementWidth - halfWindowWidth * 2.0;
                    this.Top = GetVerticalPosition(subPlacement, cy, halfElementHeight, halfWindowHeight);
                    break;
                case Placement.Top:
                    this.Left = GetHorizontalPosition(subPlacement, cx, halfElementWidth, halfWindowWidth);
                    this.Top = cy - halfElementHeight;
                    break;
                case Placement.Bottom:
                    this.Left = GetHorizontalPosition(subPlacement, cx, halfElementWidth, halfWindowWidth);
                    this.Top = cy + halfElementHeight - halfWindowHeight * 2.0;
                    break;
                case Placement.Center:
                    this.Left = cx - halfWindowWidth;
                    this.Top = cy - halfWindowHeight;
                    break;
                default:
                    this.Left = cx + halfElementWidth;
                    this.Top = cy + halfElementHeight;
                    break;
            }
        }

        /// <summary>
        /// Рассчитать уточнённое положение по вертикали.
        /// </summary>
        /// <param name="placement">Положение.</param>
        /// <param name="cy">Центр элемента по оси у.</param>
        /// <param name="halfElementHeight">Половина высоты элемента.</param>
        /// <param name="halfWindowHeight">Половина высоты окна.</param>
        /// <returns>Позиция по вертикали.</returns>
        private static double GetVerticalPosition(
            ArrowPlacement placement,
            double cy,
            double halfElementHeight,
            double halfWindowHeight) =>
           placement switch
           {
               ArrowPlacement.Top => cy - halfElementHeight,
               ArrowPlacement.Bottom => cy + halfElementHeight - halfWindowHeight * 2,
               _ => cy - halfWindowHeight,
           };

        /// <summary>
        /// Рассчитать уточнённое положение по горизонтали.
        /// </summary>
        /// <param name="placement">Положение.</param>
        /// <param name="cx">Центр элемента по оси х.</param>
        /// <param name="halfElementWidth">Половина ширины элемента.</param>
        /// <param name="halfWindowWidth">Половина ширины окна.</param>
        /// <returns>Позиция по горизонтали.</returns>
        private static double GetHorizontalPosition(
            ArrowPlacement placement,
            double cx,
            double halfElementWidth,
            double halfWindowWidth) =>
            placement switch
            {
                ArrowPlacement.Left => cx - halfElementWidth,
                ArrowPlacement.Right => cx + halfElementWidth - halfWindowWidth * 2,
                _ => cx - halfWindowWidth,
            };


        private void MenuButtonClick(object sender, RoutedEventArgs e) =>
            this.SetState(TrainingControlState.ShowMenu);

        private void ExitButtonClick(object sender, RoutedEventArgs e) =>
            this.SetState(TrainingControlState.Exit);

        /// <summary>
        /// Установить состояние окна и просигнализировать об этом
        /// для отмены текущей асинхронной операции и обработки состояния окна.
        /// </summary>
        /// <param name="state">Устанавливаемое состояние.</param>
        private void SetState(TrainingControlState state)
        {
            this.state = state;
            this.cancellationTokenSource.Cancel();
        }

        #endregion
    }
}
